/* --------------------------------------------------------------  */
/* (C)Copyright 2007                                               */
/* International Business Machines Corporation                     */
/* All Rights Reserved.                                            */
/*                                                                 */
/* Redistribution and use in source and binary forms, with or      */
/* without modification, are permitted provided that the           */
/* following conditions are met:                                   */
/*                                                                 */
/* Redistributions of source code must retain the above copyright  */
/* notice, this list of conditions and the following disclaimer.   */
/*                                                                 */
/* Redistributions in binary form must reproduce the above         */
/* copyright notice, this list of conditions and the following     */
/* disclaimer in the documentation and/or other materials          */
/* provided with the distribution.                                 */
/*                                                                 */
/* Neither the name of IBM Corporation nor the names of its        */
/* contributors may be used to endorse or promote products         */
/* derived from this software without specific prior written       */
/* permission.                                                     */
/*                                                                 */
/* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND          */
/* CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,     */
/* INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF        */
/* MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE        */
/* DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR            */
/* CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,    */
/* SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT    */
/* NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;    */
/* LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)        */
/* HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN       */
/* CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR    */
/* OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,  */
/* EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.              */
/* --------------------------------------------------------------  */

#ifndef __ALF_AAL_H
#define __ALF_AAL_H

#include "alf_def.h"
#include "alf_common.h"

/**
 * The command from PPU
 * _ALF_CMD_EXIT: The SPE thread should exit
 * _ALF_CMD_TASK_CHANGED: The task executing by SPE thread is changed.
 * _ALF_CMD_NONE: when SPE thread sees _ALF_CMD_TASK_CHANGED, it should write back _ALF_CMD_NONE
 *                to PPU to indicate it has processed this command 
 */
#define _ALF_CMD_EXIT         0xDEAD
#define _ALF_CMD_START_ACK    0xEADC
#define _ALF_CMD_EXIT_ACK     0xAECD
#define _ALF_CMD_TASK_CHANGED 0xCAED
#define _ALF_CMD_NONE         0x0000

/**
 * Status of work block 
 * _ALF_WORK_BLOCK_STATUS_INIT: The work block has schduled to execute.
 * _ALF_WORK_BLOCK_STATUS_EQ:   The work block has been put into queue.
 * _ALF_WORK_BLOCK_STATUS_FIN:  The work block has been executed successfully.
 *
 */
#define _ALF_WORK_BLOCK_STATUS_INIT   0x1
#define _ALF_WORK_BLOCK_STATUS_EQ     0x2
#define _ALF_WORK_BLOCK_STATUS_FIN    0x3

/**
 * Kind of request packet.
 * _ALF_REQ_PKT_WBH:       the request packet is a work block header
 * _ALF_REQ_PKT_SWAP_CTX:  the request packet is to swap context
 * _ALF_REQ_PKT_MERGE_CTX: the request packet is to merge context
 */
#define _ALF_REQ_PKT_WBH        0x0
#define _ALF_REQ_PKT_SWAP_CTX   0x1
#define _ALF_REQ_PKT_MERGE_CTX  0x2
#define _ALF_REQ_PKT_KINDS      0x4

/** Length for each SPE request packet queue */
#define _ALF_ACTUAL_SPU_QLEN          3
#define _ALF_SPU_QLEN                 (_ALF_ACTUAL_SPU_QLEN  + 1)


//! Size of a single dma entry
#define _ALF_MFC_ELEMENT_SIZE     8

//! Maxim dma list number supported in a single data transfer information structure
#define _ALF_MAX_DMALIST_NUM      8

//! Maxim data transfer size for a single dma entry
#define _ALF_MAX_SINGLE_DT_SIZE (16 * 1024)

//! Maxim number of dma entries can be supported in a single dma list
#define _ALF_DMA_LIST_MAX_ELEMENT_NUM   (2 * 1024)


#define _ALF_DMA_DEFAULT_ELEMENT_NUM   (128)


//! Maxim spe local store buffer size for working block header, user data and etc.
#define _ALF_MAX_SPE_LS_SIZE_ 240 * 1024

//! Maxim alignment factor which is allowed
#define _ALF_MAX_ALIGN_FACTOR_  8

//! Maxim dma entry number
#define _ALF_MAX_DMA_ENTRY_NUM  (_ALF_MAX_DMALIST_NUM * _ALF_DMA_LIST_MAX_ELEMENT_NUM)


#pragma pack(4)

typedef struct _alf_error_msg 
{
  int error_code;
  ALF_ERR_TYPE_T error_type;
  int extra_error_code;
  ALF_ERR_POLICY_T error_handle_policy;
  unsigned char _pad[128 - 4 * 4];
} alf_error_msg;

/**
 * Data entry descriptor. 
 * It will describe the constituion of a continuous memory,
 * Data entry is divided into multiple units.
 */
typedef struct _entry_desc
{
  ALF_DATA_TYPE_T type; //unit's data type
  int size; //how many units
} entry_desc_t;

/**
 * Task information, holds all the attributes of a task which will be
 * used by SPE thread to do memory allocation.
 */
typedef struct _alf_task_info 
{
  unsigned int accels;            //number of accelerators executing this task
  unsigned int ctx_size;          //task context size
  alf_data_addr64_t ctx_entry_desc_ea; //effective address of context data entry descriptor, please see entry_desc_t
  unsigned int ctx_entry_num;     //number of context data entry descriptors
  unsigned int wb_max_udata_size; //maximum user data size in work block
  unsigned int wb_max_in_size;    //maximum input data size in work block
  unsigned int wb_max_out_size;   //maximum output data size in work block
  unsigned int wb_max_inout_size; //maximum inout data size in work block
  unsigned int wb_max_dt_size;    //maximum data transfer size in work block
  unsigned int wb_max_stack_size; //maximum stack size
  unsigned int wb_max_edesc_size; //maximum entry descriptor size for work block
  unsigned int task_attr;         //task attr
  alf_data_addr64_t task_handle_ea;    //effective address of task handle  
  unsigned int api_addrs[5];
  unsigned int pad[12];            //pad to let sizeof(alf_task_info_t) = 128 bytes
} alf_task_info_t;

/**
 * The top-level communication channel between PPU and SPU.
 */
typedef struct _alf_spu_instance 
{
  volatile int spu_index;        //request packet queue index modified by SPU
  volatile int ppu_index;        //request packet queue index modified by PPU
  alf_data_addr64_t pkt_queue_ea;  //effective address of request packet queue
  alf_data_addr64_t ti_ea;         //effective address of alf_task_info_t
  alf_data_addr64_t error_ea;      //effective address of error status
  int fin_pkts[_ALF_REQ_PKT_KINDS]; //fin_pkts[0]: finished work blocks by SPU
                                   //fin_pkts[1]: finished swap context operations by SPU
                                   //fin_pkts[2]: finished merge context operations by SPU
                                   //fin_pkts[3]: not use, reserved
  volatile int cmd;              //command from PPU
  int id;               //ID of this SPU
  char pad[64];         //pad to let sizeof(alf_spu_instance_t) = 128 bytes

  /**** TODO FIND OUT why the SPE crashes when dataset_ea is placed above the pad *****/
  alf_data_addr64_t dataset_ea;    //effective address of dataset if one is available.
} alf_spu_instance_t;

/**
 * Work block header, holds all the information about a work block in PPU
 */
typedef struct _alf_wbh 
{
  int status;       //work block status
  int is_spu_dma;   //if data partition on SPU
  int total_count;  //total iteration count
  int udata_size;   //user data size
  alf_data_addr64_t udata_ea;  //effective address of user data
  alf_data_addr64_t dti_ea;    //effective address of alf_data_transfer_info_t
  alf_data_addr64_t pal_wb;    //tentative keeped, SPU don't care this field.
} alf_wbh_t;

/**
 * Describe context operation.
 */
typedef struct _alf_ctx
{
  alf_data_addr64_t get_ea; //effective address of remote memory to cast out local context
  alf_data_addr64_t set_ea; //effective address of remote memory to set into local context
  alf_data_addr64_t merge_ea; //effective address of remote memory to merge into local context
} alf_ctx_t;

/**
 * Request packet, it can be a work block header, swap context or merge context.
 * PPU will put the effective address of each request packet in queue.
 */
typedef struct _alf_req_pkt
{
  union
  {
    alf_wbh_t wbh;
    alf_ctx_t ctx;
  };
  int kind;           //kind of request packet
  char pad[124 - sizeof(alf_wbh_t)];
} alf_req_pkt_t;


/**
 * Describe a MFC element required by CELL hardware.
 */
typedef struct _mfc_element
{
  unsigned int notify : 1;
  unsigned int reserved : 15;
  unsigned int size : 16;
  unsigned int eal : 32;
} mfc_element_t;

/**
 * Describe a MFC dma list.
 * SPU will use this data structure to issue a DMA Get List command.
 */
typedef struct _alf_mfc_dma_list
{
  unsigned int eah32;        //high 32-bit of effective address
  int elmnt_offset; //offset to elments_ea in alf_data_transfer_info_t
  int elmnt_cnt;    //number of MFC elements belong to this dma list
  int ls_off;       //local store offset 
  int type;                  //which type of I/O buffer this dma list targets to
  int cur_ls_size;  //the data transfer size of this dma list
} alf_mfc_dma_list_t;

#define _ALF_MAX_DTL_NUM 40
#define _ALF_MAX_SINGLE_DT_SIZE (16 * 1024)
/**
 * Describe data transfer information for a work block.
 * SPU will use this data structure to get input data and put output data.
 */
typedef struct _alf_data_transfer_info
{
  alf_mfc_dma_list_t dmalist_ptr[_ALF_MAX_DTL_NUM]; //array of dma list
  int dmalist_num;                        //number of used dma list
  int max_elmnts_num;                     //maximum number of MFC elements
  alf_data_addr64_t elmnts_ea;         //effective address of all the MFC elements
  alf_data_addr64_t entry_desc_ea;     //one-to-one relationship with elmnts_ea, i.e. for each MFC element,
                            //there is a data entry descriptor
  int elmnts_num;  //number of available MFC elements in elmnts_ea
  unsigned int lsptr_edesc; //local store address of the entry description list
} alf_data_transfer_info_t;

#ifdef _ALF_PLATFORM_HYBRID_
/**
 * Describes a dataset buffer
 * This data structure is quadword aligned
 */ 
typedef struct _alf_dataset_buffer
{
  alf_data_addr64_t host_addr;           // host address of the buffer
  alf_data_addr64_t mid_offset;            //offset of the buffer in the PPE remote shared memory. 
  unsigned long long size;               //size of the buffer
  ALF_DATASET_ACCESS_MODE_T access_mode; //access mode
  unsigned int pad;
} alf_dataset_buffer_t;
#endif /* _ALF_PLATFORM_HYBRID_ */

#pragma pack()

#endif
