/* --------------------------------------------------------------  */
/* (C)Copyright 2007                                               */
/* International Business Machines Corporation                     */
/* All Rights Reserved.                                            */
/*                                                                 */
/* Redistribution and use in source and binary forms, with or      */
/* without modification, are permitted provided that the           */
/* following conditions are met:                                   */
/*                                                                 */
/* Redistributions of source code must retain the above copyright  */
/* notice, this list of conditions and the following disclaimer.   */
/*                                                                 */
/* Redistributions in binary form must reproduce the above         */
/* copyright notice, this list of conditions and the following     */
/* disclaimer in the documentation and/or other materials          */
/* provided with the distribution.                                 */
/*                                                                 */
/* Neither the name of IBM Corporation nor the names of its        */
/* contributors may be used to endorse or promote products         */
/* derived from this software without specific prior written       */
/* permission.                                                     */
/*                                                                 */
/* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND          */
/* CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,     */
/* INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF        */
/* MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE        */
/* DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR            */
/* CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,    */
/* SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT    */
/* NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;    */
/* LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)        */
/* HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN       */
/* CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR    */
/* OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,  */
/* EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.              */
/* --------------------------------------------------------------  */

#include <stdio.h>
#include <stdlib.h>
#include <sys/stat.h>
#include <unistd.h>
#include <pthread.h>
#include <libmisc.h>
#include <string.h>
#include <assert.h>
#include <errno.h>
#include <time.h>
#include <sys/time.h>
#include <dirent.h>
#include <cbe_mfc.h>

#include <alf_misc.h>
#include "alf_pal_internal.h"
#include "alf_misc_ppu.h"
#include "alf_common.h"
#include "alf_debug.h"
#include <alf_errno.h>

//********************************************************
// APIs
//********************************************************

//--------------------------------------------------------
// Framework APIs
//--------------------------------------------------------

/**
 * Count the total system spe number
 */
int spe_count_physical_spes()
{
  int count = 0;
  DIR *dirp;
  struct dirent *dptr;
  if ((dirp = opendir("/sys/devices/system/spu")) != NULL) {
    while ((dptr = readdir(dirp))) {
      count++;
    }
    closedir(dirp);

    /* . and .. should not been counted */
    count -= 2;
  }
  return count;
}


int _alf_pal_is_task_info_changed(alf_pal_thread_t * thread_handle, alf_pal_task_info_t * task_info)
{
  alf_task_info_t *p_tinfo = &thread_handle->data.task_info;

  if (p_tinfo->wb_max_udata_size != task_info->parm_ctx_buffer_size)
    return 1;

  return 0;
}


/**
 * Init the task info for SPU runtime according alf_pal_task_info
 */
int
_alf_pal_task_info_init(alf_pal_thread_t * p_alf_pal_thread,
                        alf_pal_task_info_t * task_info, 
                        unsigned int num_of_accels, int is_reset)
{
  alf_task_info_t *p_tinfo = &p_alf_pal_thread->data.task_info;
  alf_pal_sym_info_t *p_sym;
  unsigned int total_size = 0u;
  int i;

  if (task_info->parm_ctx_buffer_size > _ALF_MAX_SPE_LS_SIZE_ ||
      task_info->task_context_size > _ALF_MAX_SPE_LS_SIZE_ ||
      task_info->input_buffer_size > _ALF_MAX_SPE_LS_SIZE_ ||
      task_info->output_buffer_size > _ALF_MAX_SPE_LS_SIZE_ ||
      task_info->overlapped_buffer_size > _ALF_MAX_SPE_LS_SIZE_ ||
      task_info->dt_list_entries > _ALF_MAX_DMA_ENTRY_NUM || task_info->stack_size > _ALF_MAX_SPE_LS_SIZE_) {
    _ALF_DPRINTF(_ALF_ERR_LEVEL_FATAL_, "Invalid io size\n");
    return -ALF_ERR_NOEXEC;
  }

  if (task_info->dt_list_entries == 0) {
    /* If 0, set the default number */
    p_tinfo->wb_max_dt_size = sizeof(alf_data_transfer_info_t) + _ALF_DMA_DEFAULT_ELEMENT_NUM * _ALF_MFC_ELEMENT_SIZE;
  } else {
    p_tinfo->wb_max_dt_size = sizeof(alf_data_transfer_info_t) + task_info->dt_list_entries * _ALF_MFC_ELEMENT_SIZE;
  }

  p_tinfo->wb_max_dt_size = _ALF_DMA_SIZE_ALIGN_(p_tinfo->wb_max_dt_size);

  p_tinfo->accels = num_of_accels;
  p_tinfo->wb_max_udata_size = task_info->parm_ctx_buffer_size;
  p_tinfo->ctx_size = task_info->task_context_size;
  p_tinfo->wb_max_in_size = task_info->input_buffer_size;
  p_tinfo->wb_max_out_size = task_info->output_buffer_size;
  p_tinfo->wb_max_inout_size = task_info->overlapped_buffer_size;
  p_tinfo->wb_max_stack_size = task_info->stack_size;
  p_tinfo->task_handle_ea = task_info->task_handle_ea;
  PTR_TO_ADDR64(task_info->context_desc, p_tinfo->ctx_entry_desc_ea);
  p_tinfo->ctx_entry_num = task_info->context_desc_num;

  //FIXME
  p_tinfo->task_attr = task_info->task_attr;

#ifdef _ALF_PLATFORM_HYBRID_
  if (task_info->dt_list_entries == 0) {
    p_tinfo->wb_max_edesc_size = _ALF_DMA_SIZE_ALIGN_(_ALF_DMA_DEFAULT_ELEMENT_NUM * sizeof(entry_desc_t));
  } else {
    p_tinfo->wb_max_edesc_size = _ALF_DMA_SIZE_ALIGN_(task_info->dt_list_entries * sizeof(entry_desc_t));
  }
#else
  p_tinfo->wb_max_edesc_size = 0;
#endif

  if( !is_reset ) {
    p_sym = (alf_pal_sym_info_t *)task_info->p_arch_tinfo;
    for(i=0; i<ALF_API_KERNEL_NUM; i++) {
      p_tinfo->api_addrs[i] = p_sym->api_addrs[i];
    }
  }
  else {
    /*  Symbol name to address when thread reset */
    if(_alf_pal_symname_to_addr(p_alf_pal_thread->data.lib_handle, p_tinfo, task_info) < 0)
      return -ALF_ERR_NOEXEC; 
  }
 
  total_size += _ALF_DMA_SIZE_ALIGN_(p_tinfo->ctx_size);
  total_size += _ALF_DMA_SIZE_ALIGN_(p_tinfo->wb_max_udata_size);
  total_size += _ALF_DMA_SIZE_ALIGN_(p_tinfo->wb_max_in_size);
  total_size += _ALF_DMA_SIZE_ALIGN_(p_tinfo->wb_max_out_size);
  total_size += _ALF_DMA_SIZE_ALIGN_(p_tinfo->wb_max_inout_size);
  total_size += _ALF_DMA_SIZE_ALIGN_(p_tinfo->wb_max_dt_size);
  total_size += _ALF_DMA_SIZE_ALIGN_(p_tinfo->wb_max_stack_size);


  if (total_size > _ALF_MAX_SPE_LS_SIZE_) {
    goto ERR;
  }

  return 0;

ERR:

  _ALF_DPRINTF(_ALF_ERR_LEVEL_FATAL_, "Invalid task information "
               "task_max_context_size %d "
               "wb_max_udata_size %d "
               "wb_max_in_size %d wb_max_out_size %d wb_max_inout_size %d"
               "wb_max_dti__size %d"
               "wb_max_stack_size %d\n",
               p_tinfo->ctx_size,
               p_tinfo->wb_max_udata_size,
               p_tinfo->wb_max_in_size,
               p_tinfo->wb_max_out_size,
               p_tinfo->wb_max_inout_size, p_tinfo->wb_max_dt_size, p_tinfo->wb_max_stack_size);

  return -ALF_ERR_NOEXEC;

}


/**
 * Get the logical id of PAL thread
 */
unsigned int _alf_pal_get_accel_id(alf_pal_thread_t * p_alf_pal_thread)
{
  alf_pal_descriptor_t *p_alf_pal_desc;
  unsigned int accel_id;

  p_alf_pal_desc = (alf_pal_descriptor_t *) (p_alf_pal_thread->data.alf_pal_desc_ptr);

#ifdef _ALF_32B_
  accel_id = ((unsigned long) p_alf_pal_thread - (unsigned long) (p_alf_pal_desc->threads)) / sizeof(alf_pal_thread_t);
#elif _ALF_64B_
  accel_id = ((unsigned long long) p_alf_pal_thread -
              (unsigned long long) (p_alf_pal_desc->threads)) / sizeof(alf_pal_thread_t);
#endif
  return accel_id;
}



int _alf_pal_dtl_entry_add(alf_data_transfer_info_t *dtl_info,
                           alf_pal_dtl_handle dtl_handle,
                           alf_data_addr64_t ea_remote, unsigned int size)
{
  int remain = 0;
  int max_transfers = 0;
  int i;
  int ret;

  max_transfers = size / _ALF_MAX_SINGLE_DT_SIZE;
  remain = size % _ALF_MAX_SINGLE_DT_SIZE;

  for (i = 0; i < max_transfers; ++i) {

    if ((ret = _alf_pal_dtlist_add_entry(dtl_info, dtl_handle, _ALF_MAX_SINGLE_DT_SIZE, ALF_DATA_BYTE, ea_remote)) < 0) {
      _ALF_DPRINTF(_ALF_ERR_LEVEL_FATAL_, "Add dtl entry failed\n");
      return ret;
    }
    ea_remote = ea_remote + _ALF_MAX_SINGLE_DT_SIZE;
  }

  if (remain) {

    if ((ret = _alf_pal_dtlist_add_entry(dtl_info, dtl_handle, remain, ALF_DATA_BYTE, ea_remote)) < 0) {
      _ALF_DPRINTF(_ALF_ERR_LEVEL_FATAL_, "Add dtl entry failed\n");
      return ret;
    }
  }

  return 0;
}

/**
 * DTL entry add routine
 *   Add one entry in dtl list, if this dt list reach the max number or accross the 4G border,
 *   this list will be splited into two
 */
int _alf_pal_dtlist_add_entry(alf_data_transfer_info_t * dtl_info,
                              unsigned dtl_index, int chunk_size, int data_type, alf_data_addr64_t ea_remote)
{
  unsigned int new_dma_elements_size = 0;
  mfc_element_t *dma_elmnt_ptr = NULL;
  mfc_element_t *elmnt_ptr = NULL;
  mfc_element_t *new_dma_elmnt_ptr = NULL;
  int dt_size = data_type & 0xf;
  int chunk_bytes = 0;
  int across_border = 0;

  chunk_bytes = dt_size * (chunk_size);


  ADDR64_TO_PTR(dtl_info->elmnts_ea, mfc_element_t *, dma_elmnt_ptr);

  /* DMA elements buffer full, need realloc */
  if (dtl_info->elmnts_num >= dtl_info->max_elmnts_num) {

    new_dma_elements_size = (dtl_info->max_elmnts_num + 16) * _ALF_MFC_ELEMENT_SIZE;
    new_dma_elmnt_ptr = malloc_align(new_dma_elements_size, 7);
//    new_dma_elmnt_ptr = realloc_align(dma_elmnt_ptr, new_dma_elements_size, 7);
    if (new_dma_elmnt_ptr == NULL) {
      _ALF_DPRINTF(_ALF_ERR_LEVEL_FATAL_, "realloc failed\n");
      return -ALF_ERR_NOMEM;
    }

    memcpy(new_dma_elmnt_ptr, dma_elmnt_ptr, dtl_info->max_elmnts_num * _ALF_MFC_ELEMENT_SIZE);

    free_align(dma_elmnt_ptr);

    /* Set the new address */
    PTR_TO_ADDR64(new_dma_elmnt_ptr, dtl_info->elmnts_ea);

    dtl_info->max_elmnts_num += 16;
    dma_elmnt_ptr = new_dma_elmnt_ptr;
  }


  /* Check dma list reach the max elements number or not or the address across the 4G border, 
   * split it into two
   */
  do {

    across_border = ((dtl_info->dmalist_ptr[dtl_index].eah32 != (unsigned int) (ea_remote >> 32))
                     && (dtl_info->dmalist_ptr[dtl_index].elmnt_cnt > 0));


    if (dtl_info->dmalist_ptr[dtl_index].elmnt_cnt >= _ALF_DMA_LIST_MAX_ELEMENT_NUM || across_border) {

      if (dtl_index >= _ALF_MAX_DTL_NUM)
        return -ALF_ERR_NOMEM;

      /* Move to next dma list */
      if (dtl_info->dmalist_ptr[dtl_index + 1].elmnt_cnt == 0) {
        /* split into two, init the DMA list */
        dtl_info->dmalist_ptr[dtl_index + 1].eah32 = (unsigned int) (ea_remote >> 32);
        dtl_info->dmalist_ptr[dtl_index + 1].elmnt_offset = dtl_info->elmnts_num;
        dtl_info->dmalist_ptr[dtl_index + 1].elmnt_cnt = 0;
        dtl_info->dmalist_ptr[dtl_index + 1].ls_off = dtl_info->dmalist_ptr[dtl_index].ls_off +
            dtl_info->dmalist_ptr[dtl_index].cur_ls_size;
        dtl_info->dmalist_ptr[dtl_index + 1].cur_ls_size = 0;
        dtl_info->dmalist_ptr[dtl_index + 1].type = dtl_info->dmalist_ptr[dtl_index].type;
        dtl_index++;
        dtl_info->dmalist_num++;
      } else {
        /*Already splited at the privous call, no need init again */
        dtl_index++;
      }

      /* check the index is reach the limit or not */
      if (dtl_index >= _ALF_MAX_DTL_NUM)
        return -ALF_ERR_NOMEM;

    } else {
      /* find the coreect dtl_index */
      break;
    }

  } while (1);


  /* Add the element in the DMA elements buffer, update dma list info and 
   * dtl_info
   */
  elmnt_ptr = (mfc_element_t *) (dma_elmnt_ptr + dtl_info->elmnts_num);
  elmnt_ptr->notify = 0;
  elmnt_ptr->reserved = 0;
  elmnt_ptr->size = chunk_bytes;
  elmnt_ptr->eal = (unsigned int) ea_remote;

#ifdef _ALF_64B_
  if( dtl_info->dmalist_ptr[dtl_index].elmnt_cnt == 0) {
    dtl_info->dmalist_ptr[dtl_index].eah32 = (unsigned int) (ea_remote >> 32);
  }
#endif

  dtl_info->dmalist_ptr[dtl_index].elmnt_cnt++;
  dtl_info->dmalist_ptr[dtl_index].cur_ls_size += _ALF_SIMD_SIZE_ALIGN_(chunk_bytes);

  dtl_info->elmnts_num++;
  return 0;

}


/* Update all DTLs with I/O type */
void _alf_pal_io_type_combine(alf_pal_wb_t * wb)
{
  unsigned int i;
  unsigned int index;
  alf_data_transfer_info_t *dtl_info;

  dtl_info = (alf_data_transfer_info_t *) wb->dtl_grp;

  for (i = 0; i < wb->input_dtl_num; i++) {
    index = wb->input_dtls[i];
    dtl_info->dmalist_ptr[index].type = ALF_BUF_IN;
  }

  for (i = 0; i < wb->output_dtl_num; i++) {
    index = wb->output_dtls[i];
    dtl_info->dmalist_ptr[index].type = ALF_BUF_OUT;
  }

  for (i = 0; i < wb->inout_in_dtl_num; i++) {
    index = wb->inout_in_dtls[i];
    dtl_info->dmalist_ptr[index].type = ALF_BUF_OVL_IN;
  }

  for (i = 0; i < wb->inout_out_dtl_num; i++) {
    index = wb->inout_out_dtls[i];
    dtl_info->dmalist_ptr[index].type = ALF_BUF_OVL_OUT;
  }

  for (i = 0; i < wb->inout_inout_dtl_num; i++) {
    index = wb->inout_inout_dtls[i];
    dtl_info->dmalist_ptr[index].type = ALF_BUF_OVL_INOUT;
  }

  return;
}


int _alf_pal_freedq_full(alf_wq_free_dq_t * fdb_ptr)
{
  return ((fdb_ptr->rear_index + 1) % fdb_ptr->capacity == fdb_ptr->front_index);
}


/**
 * Free context request packets list create routine
 *   When thread create, one context requests packet list will be created, 
 *   a context operation will get a packet from this list 
 */
int _alf_pal_free_ctx_pkts_create(alf_pal_thread_t * thread_handle)
{
  alf_wq_free_dq_t *fdb_ptr = &thread_handle->data.freedq;
  alf_req_pkt_t *req_pkts;

  req_pkts = (alf_req_pkt_t *) malloc_align(_ALF_SPU_QLEN * sizeof(alf_req_pkt_t), 7);
  if (req_pkts == NULL) {
    return -ALF_ERR_NOMEM;
  }

  fdb_ptr->front_index = 0;
  fdb_ptr->rear_index = 0;
  fdb_ptr->capacity = _ALF_SPU_QLEN;
  thread_handle->data.ctx_reqs = req_pkts;

  return 0;
}



/**
 * Get one free context packet from the list
 */
alf_req_pkt_t *_alf_pal_free_ctx_pkt_get(alf_pal_thread_t * thread_handle)
{
  int index;
  alf_wq_free_dq_t *fdb_ptr = &thread_handle->data.freedq;
  alf_req_pkt_t *req_pkt;
  alf_ctx_t *ctx_ptr;

  if (alf_unlikely(_alf_pal_freedq_full(fdb_ptr))) {

    /* if return value <= 0, no slot can be cleared */
    if (_alf_pal_free_ctx_pkts_clear(thread_handle) <= 0)
      return NULL;
  }


  index = fdb_ptr->rear_index;
  fdb_ptr->rear_index = (index + 1) % fdb_ptr->capacity;

  req_pkt = (alf_req_pkt_t *) (thread_handle->data.ctx_reqs + index);
  ctx_ptr = (alf_ctx_t *) req_pkt;
  ctx_ptr->set_ea = 0LL;
  ctx_ptr->get_ea = 0LL;
  ctx_ptr->merge_ea = 0LL;

  return req_pkt;
}


/**
 * Check if there is some context request can be return to the list,
 * return the number of context requests returned.
 */
int _alf_pal_free_ctx_pkts_clear(alf_pal_thread_t * thread_handle)
{
  int i, j;
  alf_wq_free_dq_t *fdb_ptr = &thread_handle->data.freedq;
  alf_spu_instance_t *spu_instance = &thread_handle->data.spu_instance;
  int finished_ctx_req;

  int front = fdb_ptr->front_index;
  int end = fdb_ptr->rear_index;
  int max;

  finished_ctx_req = spu_instance->fin_pkts[_ALF_REQ_PKT_SWAP_CTX] + 
                       spu_instance->fin_pkts[_ALF_REQ_PKT_MERGE_CTX];

  max = finished_ctx_req - thread_handle->data.finished_ctx_req;
  if (max == 0) {
    /* No context request finished recently */
    return 0;
  } else if (max < 0) {
    return -ALF_ERR_GENERIC;
  }

  fdb_ptr = &thread_handle->data.freedq;
  j = 0;
  for (i = front; i != end; i = (i + 1) % fdb_ptr->capacity) {

    thread_handle->data.finished_ctx_req++;
    fdb_ptr->front_index = (i + 1) % fdb_ptr->capacity;
    j++;
    if (j >= max)
      break;
  }

  return max;
}


/**
 * Destroy the free context request list
 */
int _alf_pal_free_ctx_pkts_destroy(alf_pal_thread_t * thread_handle)
{
  alf_req_pkt_t *req_pkts;

  req_pkts = thread_handle->data.ctx_reqs;
  if (req_pkts != NULL) {
    free_align(req_pkts);
  }

  return 0;
}


/**
 * Init the request packet queue, spu 
 * runtime will get the request from this queue
 */
void _alf_pal_req_queue_init(alf_pal_thread_t * thread_handle, unsigned int spu_id)
{
  alf_spu_instance_t *spu_instance;

  spu_instance = &(thread_handle->data.spu_instance);

  spu_instance->ppu_index = spu_instance->spu_index = 0;

  spu_instance->id = spu_id;

  PTR_TO_ADDR64(&(thread_handle->data.error_handle), spu_instance->error_ea);
  PTR_TO_ADDR64(&(thread_handle->data.pkt_ea_list[0]), spu_instance->pkt_queue_ea);
  PTR_TO_ADDR64(&(thread_handle->data.task_info), spu_instance->ti_ea);
  PTR_TO_ADDR64((void*)0, spu_instance->dataset_ea);

  spu_instance->fin_pkts[_ALF_REQ_PKT_WBH] = 0;
  spu_instance->fin_pkts[_ALF_REQ_PKT_SWAP_CTX] = 0;
  spu_instance->fin_pkts[_ALF_REQ_PKT_MERGE_CTX] = 0;

  spu_instance->cmd = 0;

  thread_handle->data.error_handle.error_code = 0;
}


void _alf_pal_req_queue_reinit(alf_pal_thread_t * thread_handle)
{
  alf_spu_instance_t *spu_instance;

  spu_instance = &(thread_handle->data.spu_instance);

  spu_instance->fin_pkts[_ALF_REQ_PKT_WBH] = 0;
  spu_instance->fin_pkts[_ALF_REQ_PKT_SWAP_CTX] = 0;
  spu_instance->fin_pkts[_ALF_REQ_PKT_MERGE_CTX] = 0;

  thread_handle->data.error_handle.error_code = 0;
}


/**
 * Clear the request packet queue
 */
void _alf_pal_req_queue_clear(alf_pal_thread_t * p_alf_pal_thread)
{
  int j, max;
  int start, end;
  alf_pal_thread_t *thread_handle;
  alf_spu_instance_t *spu_instance;

  thread_handle = (alf_pal_thread_t *) p_alf_pal_thread;
  spu_instance = &thread_handle->data.spu_instance;

  start = spu_instance->spu_index;
  end = spu_instance->ppu_index;
  max = _ALF_SPU_QLEN;

  for (j = start; j != end; j = (j + 1) % max) {
    thread_handle->data.pkt_ea_list[j] = 0LL;
  }

  spu_instance->ppu_index = spu_instance->spu_index = 0;

}


/**
 * Check if all the requests finished
 */
int _alf_pal_is_all_pkt_finished(alf_pal_thread_t * thread_handle)
{
  unsigned int finished_wbs;
  unsigned int finished_ctx_req;
  unsigned int total_wbs;
  unsigned int total_ctx_req;

  _alf_pal_free_ctx_pkts_clear(thread_handle);

  finished_wbs = thread_handle->data.spu_instance.fin_pkts[_ALF_REQ_PKT_WBH];
  finished_ctx_req = thread_handle->data.spu_instance.fin_pkts[_ALF_REQ_PKT_SWAP_CTX] + thread_handle->data.spu_instance.fin_pkts[_ALF_REQ_PKT_MERGE_CTX];

  total_wbs = thread_handle->data.total_wbs;
  total_ctx_req = thread_handle->data.total_ctx_req;

  return ((total_wbs == finished_wbs) && (total_ctx_req == finished_ctx_req));
}


/**
 * Request packet queue enqueue routine.
 *   request packet enqueue, then send the notification to spu runtime
 */
int _alf_pal_req_queue_enqueue(alf_pal_thread_t * thread_handle, alf_req_pkt_t * req_pkt, int type)
{
  volatile alf_spu_instance_t *spu_instance;
  int ppu_index;

  spu_instance = &thread_handle->data.spu_instance;
  ppu_index = spu_instance->ppu_index;

  PTR_TO_ADDR64(req_pkt, thread_handle->data.pkt_ea_list[ppu_index]);
  _ALF_SYNC_MEMORY();

  spu_instance->ppu_index = (ppu_index + 1) % _ALF_SPU_QLEN;

  _alf_pal_notify_spe_thread(thread_handle);

  if (type == _ALF_REQ_PKT_WBH) {
    thread_handle->data.total_wbs++;
  } else {
    thread_handle->data.total_ctx_req++;
  }

  return 0;
}


/**
 * Notify routine
 *   notify spu runtime, these is a new request comes
 */
int _alf_pal_notify_spe_thread(alf_pal_thread_t * thread_handle)
{
  spe_sig_notify_1_area_t *sig1_ps;
  alf_spe_thread spe_thread = thread_handle->data.spe_thread;

  sig1_ps = spe_ps_area_get(spe_thread.spe_context, SPE_SIG_NOTIFY_1_AREA);

  _spe_sig_notify_1_write(sig1_ps, 1);

  return 0;
}


/**
 * Check the request packet queue is full or not
 */
int _alf_pal_is_req_queue_full(alf_spu_instance_t * spu_instance)
{

  return (((spu_instance->ppu_index + 1) % _ALF_SPU_QLEN)
          == spu_instance->spu_index);
}

/**
 * Checking whether all the spus are finished.
 *  This function checks whether all the spus accomplished their work.
 */
int _alf_pal_is_all_spu_finished(alf_pal_descriptor_t * p_alf_pal_desc)
{
  int i;
  alf_spu_instance_t *spu_instance;
  int acces = p_alf_pal_desc->acces;

  for (i = 0; i < acces; ++i) {
    spu_instance = &_ALF_GET_SPU_SLOT(p_alf_pal_desc, i).spu_instance;
    if (spu_instance->ppu_index != spu_instance->spu_index) {
      return 0;
    }
  }

  return 1;
}


static void *_alf_spe_pthread(void *data)
{

  alf_spe_thread *spe_thread = (alf_spe_thread *) data;
  unsigned int flags = 0;       //SPE_RUN_USER_REGS;
  unsigned int entry = SPE_DEFAULT_ENTRY;
  int ret;

  pthread_setcanceltype(PTHREAD_CANCEL_ASYNCHRONOUS, NULL);

  // run SPE context
  do {
    ret = spe_context_run(spe_thread->spe_context, &entry, flags, 
                          spe_thread->argp, spe_thread->envp, NULL);
    if(ret < 0) {
      printf("error when spe context run\n");
    }
  } while (ret > 0); 
  // done - now exit thread
  pthread_exit(NULL);

}

/**
 * Spu thread create routine.
 *   This function creates spe thread and initilize the spu slot.
 */
int
_alf_pal_spu_thread_create(alf_pal_handle platform_handle,
                           alf_pal_thread_t * thread_handle, spe_program_handle_t * handle)
{
  int ret = 0;
  alf_spe_thread *spe_thread;
  struct timeval now;
  unsigned long long begin_time, end_time;
  int wait_time = 0;

 /**
  Make sure that the compiler will not 'optimize' the queue addresss 
  some bugs have been found about it in PPC native compiling enviorment
 */
  volatile alf_spu_instance_t *volatile spu_instance;
  spu_instance = &thread_handle->data.spu_instance;

  thread_handle->data.state = _ALF_SPU_SLOT_ACTIVE;

  /**
     Make sure that the spu can access the queue address 
     when libspe help to load it as a "parameter" for main routine entry
  */
  _ALF_SYNC_MEMORY();

  spe_thread = &thread_handle->data.spe_thread;

  ret = spe_program_load(spe_thread->spe_context, handle);
  if (ret != 0) {
    goto RET;
  }

  spe_thread->argp = (void *) spu_instance;
  spe_thread->envp = NULL;
  ret = pthread_create(&thread_handle->data.id, NULL, &_alf_spe_pthread, spe_thread);
  if (ret != 0) {
    goto RET;
  }

  /* Should wait for the spe thread actually startup */
  gettimeofday(&now, NULL);
  begin_time = now.tv_sec * 1000 + now.tv_usec / 1000;

  while(thread_handle->data.spu_instance.cmd != _ALF_CMD_START_ACK) {

    gettimeofday(&now, NULL);
    end_time = now.tv_sec * 1000 + now.tv_usec / 1000;
    wait_time = (int) (end_time - begin_time);

    /* If time_out, seems spe thread create fail */
    if(wait_time >= _ALF_PAL_THREAD_START_TIME) {
      ret = -ALF_ERR_GENERIC;
      goto RET;
    }

  }

  thread_handle->data.spu_instance.cmd = 0;

  thread_handle->data.alf_pal_desc_ptr = platform_handle;
  thread_handle->data.state = _ALF_SPU_SLOT_ACTIVE;

RET:
  if (ret != 0) {
    thread_handle->data.state = _ALF_SPU_SLOT_UNUSED;
    _ALF_DPRINTF(_ALF_ERR_LEVEL_FATAL_, "Create accelerator error\n");
  }
  return ret;
}


/**
 * Destroy all PAL threads 
 */
void _alf_pal_spu_threads_destroy(alf_pal_descriptor_t * p_alf_pal_desc)
{
  int i;
  int time_out = 0;
  pthread_t spe_id;
  struct timeval now;
  unsigned long long begin_time, end_time;
  int wait_time = 0;

  int acces = p_alf_pal_desc->acces;

  for (i = 0; i < acces; ++i) {

    spe_id = _ALF_GET_SPU_SLOT(p_alf_pal_desc, i).id;

    if (alf_likely(spe_id != 0L)) {
      time_out = 0;
      gettimeofday(&now, NULL);
      begin_time = now.tv_sec * 1000 + now.tv_usec / 1000;

      while( _ALF_GET_SPU_SLOT(p_alf_pal_desc, i).spu_instance.cmd != _ALF_CMD_EXIT_ACK) {
        gettimeofday(&now, NULL);
        end_time = now.tv_sec * 1000 + now.tv_usec / 1000;
        wait_time = (int) (end_time - begin_time);
        if(wait_time >= _ALF_PAL_THREAD_EXIT_TIME) { 
          time_out = 1;
          break;
        }
      }

      /* If time out, maybe spe thread is in a infinit loop, so cancel it */
      if(time_out) {
        pthread_cancel(spe_id);
      }
      pthread_join(spe_id, NULL);
      _ALF_GET_SPU_SLOT(p_alf_pal_desc, i).id = 0L;
      _ALF_GET_SPU_SLOT(p_alf_pal_desc, i).state = _ALF_SPU_SLOT_UNUSED;
      _ALF_GET_SPU_SLOT(p_alf_pal_desc, i).spu_instance.cmd = 0;
    }
  }

  p_alf_pal_desc->spe_running = 0;

}


/**
 * Destroy one PAL thread 
 */
void _alf_pal_spu_thread_destroy(alf_pal_thread_t * p_alf_pal_thread)
{
  pthread_t spe_id;
  int time_out = 0;
  struct timeval now;
  unsigned long long begin_time, end_time;
  int wait_time = 0;

  spe_id = p_alf_pal_thread->data.id;

  if (alf_likely(spe_id != 0L)) {

    time_out = 0;
    gettimeofday(&now, NULL);
    begin_time = now.tv_sec * 1000 + now.tv_usec / 1000;
    while(p_alf_pal_thread->data.spu_instance.cmd != _ALF_CMD_EXIT_ACK) {
      gettimeofday(&now, NULL);
      end_time = now.tv_sec * 1000 + now.tv_usec / 1000;
      wait_time = (int) (end_time - begin_time);
      if(wait_time >= _ALF_PAL_THREAD_EXIT_TIME) {
        time_out = 1;
        break;
      }
    }

    /* If time out, maybe spe thread is in a infinit loop, so cancel it */
    if(time_out)
      pthread_cancel(spe_id);

    pthread_join(spe_id, NULL);
    p_alf_pal_thread->data.id = 0L;
    p_alf_pal_thread->data.state = _ALF_SPU_SLOT_UNUSED;
  }

}



/**
 * Deinit all the spe related resources.
 */
void _alf_pal_deinit_spes(alf_pal_descriptor_t * p_alf_pal_desc)
{
  int i;
  int acces = p_alf_pal_desc->acces;

  for (i = 0; i < acces; ++i) {
    spe_context_destroy(_ALF_GET_SPU_SLOT(p_alf_pal_desc, i).spe_thread.spe_context);
  }

  if (p_alf_pal_desc->threads) {
    free_align(p_alf_pal_desc->threads);
    p_alf_pal_desc->threads = NULL;
  }

}

/**
 * Get the total work block processing request number in the request packet queue
 */
inline int _alf_pal_wb_req_entries(alf_pal_thread_t * thread_handle)
{
  volatile alf_spu_instance_t *spu_instance;
  alf_req_pkt_t *req_pkt = NULL;
  int start, end;
  int i;
  int pending_wbs = 0;

  spu_instance = &thread_handle->data.spu_instance;

  start = spu_instance->spu_index;
  end = spu_instance->ppu_index;


  for (i = start; i != end; i = (i + 1) % _ALF_SPU_QLEN) {

    ADDR64_TO_PTR(thread_handle->data.pkt_ea_list[start], alf_req_pkt_t *, req_pkt);
    if (req_pkt->kind == _ALF_REQ_PKT_WBH)
      pending_wbs++;
  }

  return pending_wbs;
}


/**
 * Get the free entries number in the request packet queue
 */
inline int _alf_pal_req_queue_free_entries(alf_pal_thread_t * thread_handle)
{
  alf_spu_instance_t *q;
  int free_entries;
  int used_entries;

  q = &thread_handle->data.spu_instance;
  used_entries = (q->ppu_index - q->spu_index + _ALF_SPU_QLEN) % _ALF_SPU_QLEN;

  free_entries = _ALF_ACTUAL_SPU_QLEN - used_entries;

  return free_entries;
}



/**
 * Request packet init routine
 *   Init all types of request packet here 
 */
int _alf_pal_req_pkt_init(alf_pal_thread_t * thread_handle,
                          alf_req_pkt_t * pkt_ptr,
                          alf_pal_wb_t * wb, alf_data_addr64_t context_to_set,
                          alf_data_addr64_t context_to_get, alf_data_addr64_t context_to_merge, int type)
{
  switch (type) {
  case _ALF_REQ_PKT_WBH:
    {
      alf_wbh_t *wbh_ptr;
      wbh_ptr = (alf_wbh_t *) pkt_ptr;
      wbh_ptr->status = _ALF_WORK_BLOCK_STATUS_INIT;

      wbh_ptr->udata_ea = wb->parm_addr;
      wbh_ptr->udata_size = wb->parm_size;
      wbh_ptr->total_count = wb->total_count;

      //FIXME check with Interface
      wbh_ptr->is_spu_dma = (int) thread_handle->data.task_info.task_attr;

      if (!wbh_ptr->is_spu_dma) {
        //dtl_grp_handle point to the begin address of alf_data_transfer_into_t
        PTR_TO_ADDR64(wb->dtl_grp, wbh_ptr->dti_ea);
      }

      _alf_pal_io_type_combine(wb);

      // Put PAL Workblock address in workblock header
      PTR_TO_ADDR64(wb, wbh_ptr->pal_wb);

      pkt_ptr->kind = _ALF_REQ_PKT_WBH;
    }
    break;

  case _ALF_REQ_PKT_SWAP_CTX:
    {
      alf_ctx_t *ctx_req = (alf_ctx_t *) pkt_ptr;
      if ((context_to_set == 0ll) && (context_to_get == 0ll))
        return -ALF_ERR_BADF;

      ctx_req->get_ea = context_to_get;
      ctx_req->set_ea = context_to_set;

      pkt_ptr->kind = _ALF_REQ_PKT_SWAP_CTX;
    }

    break;

  case _ALF_REQ_PKT_MERGE_CTX:
    {
      alf_ctx_t *ctx_req = (alf_ctx_t *) pkt_ptr;
      if (context_to_merge == 0ll)
        return -ALF_ERR_BADF;

      ctx_req->merge_ea = context_to_merge;

      pkt_ptr->kind = _ALF_REQ_PKT_MERGE_CTX;
    }

    break;

  default:
    return -ALF_ERR_GENERIC;
    break;
  }

  return 0;
}


/**
 * Symbol name to address, used when thread reset, when reset, we only 
 * need translate the symbol name again, because maybe kernel changed
 */
int _alf_pal_symname_to_addr(void *lib_handle, alf_task_info_t *p_tinfo, 
                             alf_pal_task_info_t *task_info)
{
  char *error;
  int i;

  /**
   * Find the symbol address, max api number in each kernel is
   *  ALF_API_KERNEL_NUM
   */
  for(i=0; i<ALF_API_KERNEL_NUM; i++) {
    if(task_info->api_str[i][0]) {
      char api_str[2 * ALF_STRING_TOKEN_MAX + 2];
      strcpy(api_str, task_info->api_str[ALF_API_KERNEL_IMAGE]);
      strcat(api_str, "_");
      strcat(api_str, task_info->api_str[i]);
      api_str[2*ALF_STRING_TOKEN_MAX + 1] = 0;

      _ALF_DPRINTF(_ALF_ERR_LEVEL_TRACE_, "unsigned int *val = (unsigned int *) dlsym(lib_handle, \"%s\")\n",api_str);
      unsigned int *val = (unsigned int *) dlsym(lib_handle, api_str);
      if((error = dlerror()) != NULL)
      {
        return -ALF_ERR_NOEXEC;
      }
      p_tinfo->api_addrs[i] = *val;

    }
    else
      p_tinfo->api_addrs[i] = 0;
  }

  /* Comp kernel could not be null, this is not allowed */
  if(p_tinfo->api_addrs[ALF_API_KERNEL_COMP_KERNEL] == 0)
  {
    return -ALF_ERR_NOEXEC;
  }

  return 0;
}

 
/**
 * Symbol name to address  
 */
int _alf_pal_dlsym(alf_pal_descriptor_t *p_alf_pal_desc, alf_pal_task_info_t *task_info)
{
  spe_program_handle_t *spe_task_image;
  char fname[2 * ALF_STRING_TOKEN_MAX + 2] = { 0 };
  void *lib_handle;
  alf_pal_sym_info_t *p_sym;
  char *error;
  int ret, i;

  /* Check if dynamic library opened by other thread, if so no need to redo such thing */
  if (task_info->p_arch_tinfo != NULL ) 
  {
    if(task_info->size_of_arch_tinfo != sizeof(alf_pal_sym_info_t))
    { 
      _ALF_DPRINTF(_ALF_ERR_LEVEL_FATAL_, "Library previously opened, but sizeof task info (%d) != sizeof alf_pal_sym_info_t (%d)",task_info->size_of_arch_tinfo,sizeof(alf_pal_sym_info_t));
      return -ALF_ERR_NOEXEC;
    }
    else
    {
      return 0;
    }
  }


  /* First open the dynamic library 
   *  (1) open the dynamcic library, find the spe image
   *  (2) find the symbol address in this image 
   */
  strcpy(fname, p_alf_pal_desc->config_parm);
  strcat(fname, "/");
  strcat(fname, task_info->api_str[ALF_API_KERNEL_LIBRARY]);
  fname[2*ALF_STRING_TOKEN_MAX +1] = 0;

  _ALF_DPRINTF(_ALF_ERR_LEVEL_TRACE_, "lib_handle = dlopen(\"%s\", RTLD_LOCAL | RTLD_NODELETE | RTLD_NOW)\n",fname);
  lib_handle = dlopen(fname, RTLD_LOCAL | RTLD_NODELETE | RTLD_NOW);
  if (lib_handle == NULL) {
    _ALF_DPRINTF(_ALF_ERR_LEVEL_FATAL_, "Failed to open %s\n",fname);
    return -ALF_ERR_NOEXEC;
  }

  spe_task_image = dlsym(lib_handle, task_info->api_str[ALF_API_KERNEL_IMAGE]);
  if (spe_task_image == NULL) {
    _ALF_DPRINTF(_ALF_ERR_LEVEL_FATAL_, "Invalid task info handle: %s\n",task_info->api_str[ALF_API_KERNEL_IMAGE]);
    dlclose(lib_handle);
    return -ALF_ERR_NOEXEC;
  }

  p_sym = malloc(sizeof(alf_pal_sym_info_t));
  if(p_sym == NULL) 
  {
    _ALF_DPRINTF(_ALF_ERR_LEVEL_FATAL_, "Failed to allocate memory\n");
    return -ALF_ERR_NOMEM;
  }


  /**
   * Find the symbol address, max api number in each kernel is 
   *  ALF_API_KERNEL_NUM
   */
  for(i=0; i<ALF_API_KERNEL_NUM; i++) {
    if(task_info->api_str[i][0]) {
      char api_str[2 * ALF_STRING_TOKEN_MAX + 2];
      strcpy(api_str, task_info->api_str[ALF_API_KERNEL_IMAGE]);
      strcat(api_str, "_");
      strcat(api_str, task_info->api_str[i]);
      api_str[2*ALF_STRING_TOKEN_MAX + 1] = 0;

      _ALF_DPRINTF(_ALF_ERR_LEVEL_TRACE_, "unsigned int *val = (unsigned int *) dlsym(lib_handle, \"%s\")\n",api_str);
      unsigned int *val = (unsigned int *) dlsym(lib_handle, api_str);
      if((error = dlerror()) != NULL)
      {
	_ALF_DPRINTF(_ALF_ERR_LEVEL_FATAL_, "Failed to locate function: %s, native error: %s\n",api_str, error);
        ret = -ALF_ERR_NOEXEC;
        goto ERR;
      }
      p_sym->api_addrs[i] = *val;

    }
    else
      p_sym->api_addrs[i] = 0;
  }

  /* Comp kernel could not be null, this is not allowed */
  if(p_sym->api_addrs[ALF_API_KERNEL_COMP_KERNEL] == 0)
  {
    _ALF_DPRINTF(_ALF_ERR_LEVEL_FATAL_, "NULL computation kernel not allowed\n");
    ret = -ALF_ERR_NOEXEC;
    goto ERR;
  }

  p_sym->spe_task_image = spe_task_image;
  p_sym->lib_handle = lib_handle;
  task_info->p_arch_tinfo = p_sym;
  task_info->size_of_arch_tinfo = sizeof(alf_pal_sym_info_t);

  return 0;

ERR:
  free(p_sym);
  task_info->size_of_arch_tinfo = 0;
  return ret;
}


void _alf_report_invalid_dma()
{
  _ALF_DPRINTF(_ALF_ERR_LEVEL_FATAL_, "Invalid DMA\n");
}


/**
 * ALF PAL runtime handle checking routine.
 */
int _alf_pal_check_handle(alf_pal_descriptor_t * p_alf_pal_desc)
{
  if (!ALF_VALID_HANDLE(p_alf_pal_desc)) {
    return -EBADF;
  }

  if (p_alf_pal_desc->magic_id != _ALF_HANDLE_T_MAGIC_ID) {
    return -EBADF;
  }

  return 0;
}

/**
 * Check the status of spu thread, if error happen, call error handler 
 */
int _alf_pal_check_error(alf_pal_thread_t *thread_handle)
{
  alf_pal_descriptor_t *p_alf_pal_desc = NULL;
  ALF_ERR_POLICY_T err;

  p_alf_pal_desc = thread_handle->data.alf_pal_desc_ptr;
  if((thread_handle->data.error_handle.error_code != 0) && 
                (p_alf_pal_desc->err_handler != NULL)) {
    err = p_alf_pal_desc->err_handler(p_alf_pal_desc->err_handler_ctx,
                      thread_handle->data.error_handle.error_type,
                      thread_handle->data.error_handle.error_code,
                      NULL);
    
    thread_handle->data.error_handle.error_handle_policy = err;
    thread_handle->data.error_handle.error_code = 0;
  
    return -ALF_ERR_GENERIC;
  }
  return 0;
}
