/* --------------------------------------------------------------  */
/* (C)Copyright 2007                                               */
/* International Business Machines Corporation                     */
/* All Rights Reserved.                                            */
/*                                                                 */
/* Redistribution and use in source and binary forms, with or      */
/* without modification, are permitted provided that the           */
/* following conditions are met:                                   */
/*                                                                 */
/* Redistributions of source code must retain the above copyright  */
/* notice, this list of conditions and the following disclaimer.   */
/*                                                                 */
/* Redistributions in binary form must reproduce the above         */
/* copyright notice, this list of conditions and the following     */
/* disclaimer in the documentation and/or other materials          */
/* provided with the distribution.                                 */
/*                                                                 */
/* Neither the name of IBM Corporation nor the names of its        */
/* contributors may be used to endorse or promote products         */
/* derived from this software without specific prior written       */
/* permission.                                                     */
/*                                                                 */
/* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND          */
/* CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,     */
/* INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF        */
/* MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE        */
/* DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR            */
/* CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,    */
/* SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT    */
/* NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;    */
/* LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)        */
/* HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN       */
/* CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR    */
/* OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,  */
/* EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.              */
/* --------------------------------------------------------------  */

#include <unistd.h>
#include <alf.h>
#include "alf_host.h"
#include "alf_debug.h"

ALF_ERR_POLICY_T  alf_err_pal_error_handler(void *p_context_data, ALF_ERR_TYPE_T error_type, int error_code, char *error_string)
{
  unsigned long long err;
  alf_api_t *alf_handle = p_context_data;
  alf_err_msg err_msg; 
  err_msg.error_type = error_type;
  err_msg.error_code = error_code;
  err_msg.error_string = error_string;

  while(NULL == alf_arraylist_enqueue(alf_handle->err_msg_list,(void *)&err_msg));
  //called in pal thread which may be in different thread from scheduler thread 
  //need to lock here to get an unique number and wait for it to be processed by error thread
  pthread_mutex_lock(&alf_handle->err_lock);
  err = alf_handle->err_num + 1;
  alf_handle->err_num = err;
  pthread_mutex_unlock(&alf_handle->err_lock);
  while(alf_handle->err_finished != err)
  {
    usleep(10);
  }
  
  return err_msg.ret; 
}
ALF_ERR_POLICY_T alf_err_call_error_handler(alf_api_t* alf_handle, ALF_ERR_TYPE_T error_type, int error_code, char *err_msg)
{
    ALF_ERR_POLICY_T rtn;
    //lock to avoid user register a new error handler
    pthread_mutex_lock(&alf_handle->lock); 
    switch(error_type)
    {
        case ALF_ERR_WARNING:
            rtn = alf_handle->err_handler(alf_handle->err_handler_data_ptr, error_type, error_code, err_msg);
            break;
            
        case ALF_ERR_EXCEPTION: 
            rtn = alf_handle->err_handler(alf_handle->err_handler_data_ptr, error_type, error_code, err_msg);
            if(ALF_ERR_POLICY_IGNORE == rtn)
            {
                fprintf(stderr, "User error handler required to ignore EXCEPTIONS, will do SKIP!\n");
                rtn = ALF_ERR_POLICY_SKIP;
            }
            break;

        case ALF_ERR_FATAL:     // we have no choice but abort
            rtn = alf_handle->err_handler(alf_handle->err_handler_data_ptr, error_type, error_code, err_msg);

            if(ALF_ERR_POLICY_ABORT != rtn)
            {
                fprintf(stderr, "User error handler required to continue with FATAL, will do ABORT!\n");
                rtn = ALF_ERR_POLICY_ABORT;
            }
            break;

         default:               // if we are called wrongly
             fprintf(stderr, "ALF runtime internal error: unexpected error_type %d \n", error_type);
             rtn = ALF_ERR_POLICY_ABORT;
             break;
    }
    pthread_mutex_unlock(&alf_handle->lock); 

    return rtn;
}
ALF_ERR_POLICY_T alf_err_error_processing(alf_api_t* alf_handle, ALF_ERR_TYPE_T error_type, int error_code, char *err_msg)
{
  ALF_ERR_POLICY_T err;
	_ALF_DPRINTF(_ALF_ERR_LEVEL_ENTRYEXIT_, "Enter: alf_handle=%p, error_type=%d, error_code=%d, err_msg=%s\n", alf_handle, error_type, error_code, err_msg);
  err = alf_err_call_error_handler(alf_handle, error_type, error_code, err_msg);
  switch(err)
  {
    case ALF_ERR_POLICY_IGNORE:
    {
    }
    break;
    case ALF_ERR_POLICY_RETRY:
    {
      _ALF_DPRINTF(_ALF_ERR_LEVEL_FATAL_, "Retry unsupported yet!\n");
    }
    break;
    case ALF_ERR_POLICY_SKIP:
    {
      _ALF_DPRINTF(_ALF_ERR_LEVEL_FATAL_, "Skip error with error code %d, error string %s\n",error_code,err_msg);
    }
    break;
    case ALF_ERR_POLICY_ABORT:
    {
      _ALF_DPRINTF(_ALF_ERR_LEVEL_ALWAYS_, "An Error with error code:%d, error string %s caught, aborting.\n",
                   error_code, err_msg);
      //return to main loop to force exit 
      //the user main thread should be still alive 
      pthread_mutex_lock(&alf_handle->lock);
      alf_handle->state = ALF_API_STATUS_EXITING;
      alf_handle->exit_policy = ALF_EXIT_POLICY_FORCE; 
      pthread_mutex_unlock(&alf_handle->lock);
    }
    break;
    default:
    {
      _ALF_DPRINTF(_ALF_ERR_LEVEL_FATAL_, "Out of policy error!\n");
    }
  }
  _ALF_DPRINTF(_ALF_ERR_LEVEL_ENTRYEXIT_, "Exit\n");
  return err;
}
void *alf_api_err_thread(void *alf_ptr)
{
  alf_api_t *alf_handle = (alf_api_t *) alf_ptr;
  alf_err_msg *msg;

  do {
      if (alf_handle->state == ALF_API_STATUS_EXITED) {
        break;
      }
      if(alf_arraylist_is_empty(alf_handle->err_msg_list)) {
        usleep(10);
        continue;
      }
      msg = alf_arraylist_dequeue(alf_handle->err_msg_list); 
      msg->ret = alf_err_error_processing(alf_handle,msg->error_type, msg->error_code, msg->error_string); 
      //update count to notify waiting pal thread
      alf_handle->err_finished++;
  }
  while (1);

  return NULL;
}
