/* --------------------------------------------------------------  */
/* (C)Copyright 2007                                               */
/* International Business Machines Corporation                     */
/* All Rights Reserved.                                            */
/*                                                                 */
/* Redistribution and use in source and binary forms, with or      */
/* without modification, are permitted provided that the           */
/* following conditions are met:                                   */
/*                                                                 */
/* Redistributions of source code must retain the above copyright  */
/* notice, this list of conditions and the following disclaimer.   */
/*                                                                 */
/* Redistributions in binary form must reproduce the above         */
/* copyright notice, this list of conditions and the following     */
/* disclaimer in the documentation and/or other materials          */
/* provided with the distribution.                                 */
/*                                                                 */
/* Neither the name of IBM Corporation nor the names of its        */
/* contributors may be used to endorse or promote products         */
/* derived from this software without specific prior written       */
/* permission.                                                     */
/*                                                                 */
/* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND          */
/* CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,     */
/* INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF        */
/* MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE        */
/* DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR            */
/* CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,    */
/* SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT    */
/* NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;    */
/* LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)        */
/* HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN       */
/* CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR    */
/* OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,  */
/* EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.              */
/* --------------------------------------------------------------  */

/*
 ************************************************************
 * ALF Host Internal Utility API
 ************************************************************
 */

#include <stddef.h>
#include <string.h>
#include <stdio.h>
#include <libmisc.h>
#include <alf.h>
#include "alf_host.h"
#include "alf_debug.h"
#include "alf_api_local.h"
/*
 ************************************************************
 * Globals
 ************************************************************
 */

/*
 ************************************************************
 * Local Globals
 ************************************************************
 */

/*
 ************************************************************
 * Local APIs
 ************************************************************
 */

//---------------------------------------------------------------------------------------------------
// alf_api_task_wb_enqueue
//---------------------------------------------------------------------------------------------------

int alf_api_task_wb_enqueue(alf_api_task_t * task_handle, alf_api_wb_t * wb_handle)
{
  if (task_handle->attr & ALF_TASK_ATTR_WB_CYCLIC) {
    //we can enqueue wb directly to accel's wb queue
    if(NULL == alf_arraylist_enqueue(task_handle->accel_wbq[task_handle->cur_accel],wb_handle))
    {
      _ALF_DPRINTF(_ALF_ERR_LEVEL_TRACE_, "wb enqueue failed\n");
      return -ALF_ERR_NOMEM;
    }

    task_handle->dist_num += 1;

    _ALF_DPRINTF(_ALF_ERR_LEVEL_TRACE_, "wb enqueued to accel %d\n", task_handle->cur_accel);
    if (task_handle->dist_num == task_handle->wb_dist_size) {
      //current wb bundle finished
      //set current accel to next accel
      task_handle->cur_accel = (task_handle->cur_accel + 1) % task_handle->num_accel_req;
      task_handle->dist_num = 0;
    }
  } else {
    //enqueue wb to a global queue and leave scheduling work to scheduler
    if(NULL == alf_arraylist_enqueue(task_handle->task_wbq, wb_handle))
    {
      _ALF_DPRINTF(_ALF_ERR_LEVEL_TRACE_, "wb enqueue failed\n");
      return -ALF_ERR_NOMEM;
    }
  }
  return 0;
}
/* try to free finished wb as much as possible
**/
void alf_sched_task_wb_try_free(alf_api_task_t * task_handle)
{
  alf_api_wb_t *wb;
  alf_pal_thread_status_t status;
  unsigned int wb_num;
  unsigned int accel_index;
  unsigned int wb_index;
  //we asumme wb is processed serially on each accel
  //so when we get new finished wb number
  //we release the same number of wb on the top of the garbage queue
  for (accel_index = 0; accel_index < task_handle->num_accels; accel_index++) {
    if (alf_pal_thread_status_query(task_handle->p_task_threads[accel_index].thread, 
                                    &status) >= 0) {
      if ((wb_num = status.finished_wbs - task_handle->p_task_threads[accel_index].finished_wbs) 
            > 0) {
        for (wb_index = 0; wb_index < wb_num; wb_index++) {
          if( (wb = alf_arraylist_dequeue(task_handle->accel_fin_wbq[accel_index])) != NULL )
          {
            alf_int_wb_handle_destroy(wb);
            alf_sched_task_wb_count_dec(task_handle);
          }
        }
        task_handle->p_task_threads[accel_index].finished_wbs = status.finished_wbs;
      }
    } else
      break;
  }
}
/* free all wb for finished or destroyed task
**/
void alf_sched_task_wb_free(alf_api_task_t * task_handle)
{
  alf_api_wb_t *wb;
  int i;
  //destroy wb in task wb queue if it has
  if (task_handle->task_wbq != NULL) {
    while ((wb = (alf_api_wb_t *) alf_arraylist_dequeue(task_handle->task_wbq)) != NULL)
      alf_int_wb_handle_destroy(wb);
  }
  //destroy wb in accel garbage wb queue if it has
  for (i = 0; i < (int) task_handle->num_accel_req; i++) {
    if (NULL != task_handle->accel_fin_wbq && NULL != task_handle->accel_fin_wbq[i])
      while ((wb = (alf_api_wb_t *) alf_arraylist_dequeue(task_handle->accel_fin_wbq[i])) != NULL) {
        alf_int_wb_handle_destroy(wb);
        //only dec count for finished wb
        alf_sched_task_wb_count_dec(task_handle);
      }
  }
}
/* destroy wb resources includes:
 * wb handle
 * pal dtl group
 * wb parm
**/
void alf_int_wb_handle_destroy(alf_api_wb_t * wb)
{
    void *parm_ptr;
    int rtn;

  if(wb == NULL)
    return;
  
  rtn = alf_pal_wb_cleanup(&wb->pal);

  if(rtn < 0)
  {
      _ALF_DPRINTF(_ALF_ERR_LEVEL_TRACE_, "wb pal wb_cleanup failed %d\n", rtn);
      // FIXME: shall we choose to continue here ?
      //    the current implementation is to continue on error, else we could leak resources
      // FIXME: we need an error callback here
  }

  ADDR64_TO_PTR(wb->pal.parm_addr,void *,parm_ptr);
  if(parm_ptr != 0)
  {
    free_align(parm_ptr);
    _ALF_DPRINTF(_ALF_ERR_LEVEL_TRACE_, "wb parm free succeed\n");
  }
  
  if(wb->pal.dtl_grp != NULL)
  {
    alf_pal_dtl_group_destroy(wb->pal.dtl_grp);
    _ALF_DPRINTF(_ALF_ERR_LEVEL_TRACE_, "wb pal dtl grp free succeed\n");
  }
  
  free_align(wb);

  _ALF_DPRINTF(_ALF_ERR_LEVEL_TRACE_, "wb handle free succeed\n");
}



int alf_sched_task_list_cleanup(alf_api_t *alf_handle,alf_arraylist_t* task_list)
{
  alf_api_task_t *task_handle;
  unsigned int len = alf_arraylist_get_length(task_list);
  unsigned int i;
 
  for(i=0;i<len;i++)
  {
    if((task_handle = alf_arraylist_dequeue(task_list)) != NULL) {
      //do not count unfinished task caused by alf_task_destroy
      if(task_handle->state != ALF_API_TASK_STATUS_DESTROY)
        alf_handle->task_unfinished++;
      _ALF_API_STD_COND_SIG(task_handle->lock, task_handle->cond, 
                              task_handle->state = ALF_API_TASK_STATUS_DESTROY);
      alf_int_task_res_destroy(task_handle);
    }
  }
  return len;
}
/* destroy all task for alf force exit
*/
int alf_sched_task_cancel(alf_api_t * alf_handle)
{
  //destroy pal thread
  alf_thread_mgr_cleanup(alf_handle);
  _ALF_DPRINTF(_ALF_ERR_LEVEL_TRACE_, "alf threadpool free succeed\n");
  //cancel all tasks in init/ready/exec list
  alf_sched_task_list_cleanup(alf_handle, alf_handle->init_task_list); 
  alf_sched_task_list_cleanup(alf_handle, alf_handle->ready_task_list); 
  alf_sched_task_list_cleanup(alf_handle, alf_handle->exec_task_list); 
  return 0;
}
/*garbage collection for finished or destroyed task
*/
int alf_sched_task_release(alf_api_t * alf_handle)
{
  alf_api_task_t *task_handle;
  while ((task_handle = alf_arraylist_dequeue(alf_handle->destroy_task_list)) != NULL) {
    if(task_handle->state != ALF_API_TASK_STATUS_FINISH)
    {
      _ALF_API_STD_COND_SIG(task_handle->lock, task_handle->cond, 
                            task_handle->state = ALF_API_TASK_STATUS_DESTROY);
    }
    alf_int_task_res_destroy(task_handle);
  }
  return 0;
}
/* destroy alf runtime resources include:
 * alf handle
 * alf pal
 * alf accel pal
 * 5 task lists
 * lock & cond
 * all task handle
**/
void alf_api_handle_destroy(alf_api_t * alf_handle)
{
  alf_api_task_t *task_handle;
  alf_api_task_info_t *task_info;
  if (NULL != alf_handle) {
    //cancel scheduler thread
    if(alf_handle->scheduler != 0 )
    {
      pthread_cancel(alf_handle->scheduler);
      pthread_join(alf_handle->scheduler,NULL);
    }
    //free all task handle
    if(alf_handle->ref_task_list != NULL)
    {
      while (NULL != (task_handle = alf_arraylist_dequeue(alf_handle->ref_task_list))) {
        pthread_mutex_destroy(&task_handle->lock);
        _ALF_DPRINTF(_ALF_ERR_LEVEL_TRACE_, "task lock free succeed\n");
        pthread_cond_destroy(&task_handle->cond);
        _ALF_DPRINTF(_ALF_ERR_LEVEL_TRACE_, "task cond free succeed\n");
        free(task_handle);
        _ALF_DPRINTF(_ALF_ERR_LEVEL_TRACE_, "task handle free succeed\n");
      }
    }
    //free all task info
    if(alf_handle->task_info_list != NULL)
    {
      while (NULL != (task_info = alf_arraylist_dequeue(alf_handle->task_info_list))) {
        alf_int_task_info_destroy(task_info);
      }
    }
    //destroy pal thread
    alf_thread_mgr_cleanup(alf_handle);
    _ALF_DPRINTF(_ALF_ERR_LEVEL_TRACE_, "alf threadpool free succeed\n");
    //release accel and pal resources
    if(alf_handle->platform_handle != NULL)
    {
      alf_pal_exit(alf_handle->platform_handle);
      _ALF_DPRINTF(_ALF_ERR_LEVEL_TRACE_, "alf pal free succeed\n");
    }
    //destroy all task lists
    alf_arraylist_destroy(alf_handle->init_task_list);
    _ALF_DPRINTF(_ALF_ERR_LEVEL_TRACE_, "alf init list free succeed\n");
    alf_arraylist_destroy(alf_handle->ready_task_list);
    _ALF_DPRINTF(_ALF_ERR_LEVEL_TRACE_, "alf ready list free succeed\n");
    alf_arraylist_destroy(alf_handle->exec_task_list);
    _ALF_DPRINTF(_ALF_ERR_LEVEL_TRACE_, "alf exec list free succeed\n");
    alf_arraylist_destroy(alf_handle->destroy_task_list);
    _ALF_DPRINTF(_ALF_ERR_LEVEL_TRACE_, "alf destroy list free succeed\n");
    alf_arraylist_destroy(alf_handle->ref_task_list);
    _ALF_DPRINTF(_ALF_ERR_LEVEL_TRACE_, "alf ref list free succeed\n");
    //destroy errro msg list
    alf_arraylist_destroy(alf_handle->err_msg_list);
    _ALF_DPRINTF(_ALF_ERR_LEVEL_TRACE_, "alf err list free succeed\n");
    //destroy lock & cond
    pthread_mutex_destroy(&alf_handle->err_lock);
    _ALF_DPRINTF(_ALF_ERR_LEVEL_TRACE_, "alf err lock free succeed\n");
    pthread_mutex_destroy(&alf_handle->cond_lock);
    _ALF_DPRINTF(_ALF_ERR_LEVEL_TRACE_, "alf cond lock free succeed\n");
    pthread_cond_destroy(&alf_handle->cond);
    _ALF_DPRINTF(_ALF_ERR_LEVEL_TRACE_, "alf cond free succeed\n");
    pthread_mutex_destroy(&alf_handle->lock);
    _ALF_DPRINTF(_ALF_ERR_LEVEL_TRACE_, "alf lock free succeed\n");
    //free alf handle
    free(alf_handle);
    _ALF_DPRINTF(_ALF_ERR_LEVEL_TRACE_, "alf handle free succeed\n");
    alf_handle = NULL;
  }

}
/* destroy task desc resources include:
 * lock
 * arch tinfo
 * pal context desc
 * task desc handle
**/
void alf_int_task_info_destroy(alf_api_task_info_t * task_info)
{
  if (NULL != task_info) {
    pthread_mutex_destroy(&task_info->lock);
    if (NULL != task_info->pal.p_arch_tinfo) {
      free(task_info->pal.p_arch_tinfo);
    }
    if (NULL != task_info->pal.context_desc)
    {
      free(task_info->pal.context_desc);
    }
    free(task_info);
    task_info = NULL;
  }

}
/* destroy alf task resources includes:
 * pal task info
 * child_tasks queue
 * task wbq
 * accel wbq
 * accel garbage wbq
 * thread array
 * lock & cond
 * * * 
 * Notes: task handle itself should be destroy in alf destroy
**/
void alf_int_task_res_destroy(alf_api_task_t * task_handle)
{
  int i;

  if (NULL == task_handle)
    return;
  //destroy task work blocks
  alf_sched_task_wb_free(task_handle);
  //free pal arch tinfo
  if( NULL != task_handle->task_info.pal.p_arch_tinfo )
  {
    free(task_handle->task_info.pal.p_arch_tinfo);
    task_handle->task_info.pal.p_arch_tinfo = NULL;
    _ALF_DPRINTF(_ALF_ERR_LEVEL_TRACE_, "task info free succeed\n");
  }
  //free context desc
  if( NULL != task_handle->task_info.pal.context_desc )
  {
    free_align(task_handle->task_info.pal.context_desc);
    task_handle->task_info.pal.context_desc = NULL;
    _ALF_DPRINTF(_ALF_ERR_LEVEL_TRACE_, "task info free succeed\n");
  }
  //free dependency list 
  if ( NULL != task_handle->child_tasks)
  {
    alf_arraylist_destroy(task_handle->child_tasks);
    task_handle->child_tasks = NULL;
    _ALF_DPRINTF(_ALF_ERR_LEVEL_TRACE_, "task child list free succeed\n");
  }
  //destroy task wbq 
  if ( NULL != task_handle->task_wbq)
  {
    alf_arraylist_destroy(task_handle->task_wbq);
    task_handle->task_wbq = NULL;
    _ALF_DPRINTF(_ALF_ERR_LEVEL_TRACE_, "task wbq free succeed\n");
  }
  //free each accel wbq and accel wbq ptr array 
  if (NULL != task_handle->accel_wbq) {
    for (i = 0; i < (int) task_handle->num_accel_req; i++) {
      if (NULL != task_handle->accel_wbq[i])
      {
        alf_arraylist_destroy(task_handle->accel_wbq[i]);
        task_handle->accel_wbq[i] = NULL;
        _ALF_DPRINTF(_ALF_ERR_LEVEL_TRACE_, "task accel wbq free succeed\n");
      }
      else
        break;
    }
    free(task_handle->accel_wbq);
    task_handle->accel_wbq = NULL;
    _ALF_DPRINTF(_ALF_ERR_LEVEL_TRACE_, "task accel wbq ptr free succeed\n");
  }
  //free each accel garbage wbq and ptr array
  if (NULL != task_handle->accel_fin_wbq) {
    for (i = 0; i < (int) task_handle->num_accel_req; i++) {
      if (NULL != task_handle->accel_fin_wbq[i])
      {
        alf_arraylist_destroy(task_handle->accel_fin_wbq[i]);
        task_handle->accel_fin_wbq[i] = NULL;
        _ALF_DPRINTF(_ALF_ERR_LEVEL_TRACE_, "task accel wbq free succeed\n");
      }
      else
        break;
    }
    free(task_handle->accel_fin_wbq);
    task_handle->accel_fin_wbq = NULL;
    _ALF_DPRINTF(_ALF_ERR_LEVEL_TRACE_, "task accel fin wbq ptr free succeed\n");
  }
  //free task thread array
  if (NULL != task_handle->p_task_threads)
  {
    free(task_handle->p_task_threads);
    task_handle->p_task_threads = NULL;
    _ALF_DPRINTF(_ALF_ERR_LEVEL_TRACE_, "task thread ptr free succeed\n");
  }
  //free thread context buffer
  if (NULL != task_handle->p_thread_context)
  {
    free_align(task_handle->p_thread_context);
    task_handle->p_thread_context = NULL;
    _ALF_DPRINTF(_ALF_ERR_LEVEL_TRACE_, "task thread context free succeed\n");
  }
  //free event handler data
  if (NULL != task_handle->event_handler.p_data)
  {
    free(task_handle->event_handler.p_data);
    task_handle->event_handler.p_data = NULL;
    _ALF_DPRINTF(_ALF_ERR_LEVEL_TRACE_, "task event handler data free succeed\n");
  }
}


void alf_sched_task_wb_count_dec(alf_api_task_t * task_handle)
{
  pthread_mutex_lock(&task_handle->lock);
  task_handle->num_wb_pending--;
  pthread_mutex_unlock(&task_handle->lock);
}

int alf_sched_task_wbq_create(alf_api_task_t * task_handle)
{
  unsigned int i;
  //for non-fixed mapping
  //the number of accel wbq can only be decided after thread allocated
  if (!(task_handle->attr & ALF_TASK_ATTR_SCHED_FIXED))
  {
    //for bundled wb, we need to create accel wbq to buffer wb in one bundle
    if(task_handle->wb_dist_size > 1) {
      for (i = 0; i < task_handle->num_accels; i++) {
        if (NULL == task_handle->accel_wbq[i]) {
          task_handle->accel_wbq[i] = alf_arraylist_create(task_handle->wb_dist_size);
          if (NULL == task_handle->accel_wbq[i]) {
            return -ALF_ERR_NOMEM;
          }
        }
        else _ALF_DPRINTF(_ALF_ERR_LEVEL_TRACE_, "task accel wbq allocate succeed\n");
      }
    }
    //garbage queue is always needed
    for (i = 0; i < task_handle->num_accels; i++) {
      if (NULL == task_handle->accel_fin_wbq[i]) {
        task_handle->accel_fin_wbq[i] = alf_arraylist_create(_ALF_API_ACCEL_WB_QUEUE_LEN);
        if (NULL == task_handle->accel_fin_wbq[i]) {
          return -ALF_ERR_NOMEM;
        }
      }
      else _ALF_DPRINTF(_ALF_ERR_LEVEL_TRACE_, "task accel fin wbq allocate succeed\n");
    }
  }
  return 0;
}

int alf_api_task_wbq_create(alf_api_task_t * task_handle)
{
  unsigned int num_accelerators = task_handle->num_accel_req;
  unsigned int i;

  task_handle->accel_wbq = (alf_arraylist_t **) calloc(num_accelerators, sizeof(alf_arraylist_t *));

  if (NULL == task_handle->accel_wbq) {
    return -ALF_ERR_NOMEM;
  }
  else _ALF_DPRINTF(_ALF_ERR_LEVEL_TRACE_, "task accel wbq ptr allocate succeed\n");

  task_handle->accel_fin_wbq = (alf_arraylist_t **) calloc(num_accelerators, sizeof(alf_arraylist_t *));

  if (NULL == task_handle->accel_fin_wbq) {
    return -ALF_ERR_NOMEM;
  }
  else _ALF_DPRINTF(_ALF_ERR_LEVEL_TRACE_, "task accel fin wbq ptr allocate succeed\n");

  if (task_handle->attr & ALF_TASK_ATTR_WB_CYCLIC) {
    //for cyclic distribution, create accel wbq when task created
    for (i = 0; i < num_accelerators; i++) {
      task_handle->accel_wbq[i] = alf_arraylist_create(_ALF_API_ACCEL_WB_QUEUE_LEN);
      if (NULL == task_handle->accel_wbq[i]) {
        return -ALF_ERR_NOMEM;
      }
      else _ALF_DPRINTF(_ALF_ERR_LEVEL_TRACE_, "task accel wbq allocate succeed\n");
      //also create garbage queue
      task_handle->accel_fin_wbq[i] = alf_arraylist_create(_ALF_API_ACCEL_WB_QUEUE_LEN);
      if (NULL == task_handle->accel_fin_wbq[i]) {
        return -ALF_ERR_NOMEM;
      }
      else _ALF_DPRINTF(_ALF_ERR_LEVEL_TRACE_, "task fin accel wbq allocate succeed\n");
    }
  } else {
    //for non-cyclic distribution, create task wbq when task created
    task_handle->task_wbq = alf_arraylist_create(_ALF_API_ACCEL_WB_QUEUE_LEN);
    if (NULL == task_handle->task_wbq) {
      return -ALF_ERR_NOMEM;
    }
    else _ALF_DPRINTF(_ALF_ERR_LEVEL_TRACE_, "task wbq allocate succeed\n");
    if (task_handle->attr & ALF_TASK_ATTR_SCHED_FIXED) 
    {
      if( task_handle->wb_dist_size > 1) {
        //for bundled wb, we need to create accel wbq to buffer wb in one bundle
        //for fixed mapping we can create the accel wbq when task created
        //since we can know the number of accels at this time
        for (i = 0; i < num_accelerators; i++) {
          task_handle->accel_wbq[i] = alf_arraylist_create(task_handle->wb_dist_size);
          if (NULL == task_handle->accel_wbq[i]) {
            return -ALF_ERR_NOMEM;
          }
          else _ALF_DPRINTF(_ALF_ERR_LEVEL_TRACE_, "task accel wbq allocate succeed\n");
        }
      }
      //create garbage queue
      for (i = 0; i < num_accelerators; i++) {
        task_handle->accel_fin_wbq[i] = alf_arraylist_create(_ALF_API_ACCEL_WB_QUEUE_LEN);
        if (NULL == task_handle->accel_fin_wbq[i]) {
          return -ALF_ERR_NOMEM;
        }
        else _ALF_DPRINTF(_ALF_ERR_LEVEL_TRACE_, "task accel fin wbq allocate succeed\n");
      }
    }
  }
  return 0;
}

int alf_api_task_offspring_destroy(alf_api_task_t * task_handle)
{
  int child_num;
  int i;
  int ret = 0;

  alf_api_task_t *child_task;
  pthread_mutex_lock(&task_handle->lock);
  //for tasks not finished or destroyed
  if (task_handle->state <= ALF_API_TASK_STATUS_EXEC) {
    task_handle->state = ALF_API_TASK_STATUS_DESTROY;
    alf_int_task_call_event_handler(task_handle, ALF_TASK_EVENT_DESTROY);
    child_num = alf_arraylist_get_length(task_handle->child_tasks);
    for (i = 0; i < child_num; i++) {
      child_task = alf_arraylist_dequeue(task_handle->child_tasks);
      alf_api_task_offspring_destroy(child_task);
    }
  } else if (task_handle->state == ALF_API_TASK_STATUS_DESTROY ) {
    ret = -ALF_ERR_SRCH;
    _ALF_DPRINTF(_ALF_ERR_LEVEL_API_, "task already destroyed\n");
  } 
  //do nothing for finished task
  pthread_mutex_unlock(&task_handle->lock);
  return ret;
}

int alf_int_task_call_event_handler(alf_api_task_t * task_handle, ALF_TASK_EVENT_TYPE_T event)
{
  if (task_handle->event_handler.event_mask & event) {
      return (int) (*task_handle->event_handler.callback_func) (task_handle, event, task_handle->event_handler.p_data);
  }
  return 0;
}

ALF_ERR_POLICY_T  alf_err_default_error_handler(void *p_context_data  DECLARE_UNUSED, ALF_ERR_TYPE_T error_type, int error_code, char *error_string)
{
    // I need not to check parms here as you will see
    switch(error_type)
    {
        case ALF_ERR_WARNING:   // warnings shall be ignored
            if(error_string)
                fprintf(stderr, "ALF runtime warning: %d '%s'\n", error_code, error_string);
            else
                fprintf(stderr, "ALF runtime warning: %d\n", error_code);
            return ALF_ERR_POLICY_IGNORE;
            
        case ALF_ERR_EXCEPTION: // exceptions shall abort by default
            if(error_string)
                fprintf(stderr, "ALF runtime exception: %d '%s'\n", error_code, error_string);
            else
                fprintf(stderr, "ALF runtime exception: %d\n", error_code);
            return ALF_ERR_POLICY_ABORT;

        case ALF_ERR_FATAL:     // we have no choice but abort
            if(error_string)
                fprintf(stderr, "ALF runtime error: %d '%s'\n", error_code, error_string);
            else
                fprintf(stderr, "ALF runtime error: %d\n", error_code);
            return ALF_ERR_POLICY_ABORT;

         default:               // if we are called wrongly
             fprintf(stderr, "ALF runtime internal error: unexpected error_type %d \n", error_type);
             return ALF_ERR_POLICY_ABORT;
    }
    // we need no return here, zombie code
}
int alf_api_task_info_copy(alf_api_task_t* task_handle, alf_api_task_info_t* task_info, 
                       void *p_task_context_data)
{
  void *ptmp;
  /* library, image and comp kernel string could not be null, this is not allowed */
  if(!task_info->pal.api_str[ALF_API_KERNEL_COMP_KERNEL][0] ||
     !task_info->pal.api_str[ALF_API_KERNEL_LIBRARY][0] ||
     !task_info->pal.api_str[ALF_API_KERNEL_IMAGE][0] )
  {
    _ALF_DPRINTF(_ALF_ERR_LEVEL_API_, "api str is null: comp_kernel=\"%s\", library=\"%s\", image=\"%s\"\n",task_info->pal.api_str[ALF_API_KERNEL_COMP_KERNEL],task_info->pal.api_str[ALF_API_KERNEL_LIBRARY],task_info->pal.api_str[ALF_API_KERNEL_IMAGE]);
    return -ALF_ERR_NOEXEC;
  }
  //copy task info content
  memcpy(&task_handle->task_info, task_info, sizeof(alf_api_task_info_t));
  if(p_task_context_data != NULL && task_info->pal.context_desc_num != 0)
  {

    //store context ptr
    task_handle->p_context = p_task_context_data;

    //copy context desc
    ptmp = malloc_align(task_info->pal.context_desc_num*sizeof(alf_pal_data_entry_desc_t),
                        task_handle->task_info.accel_align);  
    if (ptmp == NULL) {
      _ALF_DPRINTF(_ALF_ERR_LEVEL_API_, "task ctx desc allocate memory failed\n");
      return -ALF_ERR_NOMEM;
    }
    else _ALF_DPRINTF(_ALF_ERR_LEVEL_TRACE_, "task ctx desc allocate succeed\n");
    task_handle->task_info.pal.context_desc = ptmp;

    memcpy(task_handle->task_info.pal.context_desc, task_info->pal.context_desc,
         task_info->pal.context_desc_num*sizeof(alf_pal_data_entry_desc_t));

    //each accel have a buffer for task context
    task_handle->task_info.pal.task_context_size = _ALF_SIZE_ALIGN_(
                                                   task_handle->task_info.ctx_entry_size,
                                                   task_handle->task_info.accel_align);
    task_handle->p_thread_context = (void *) malloc_align((task_handle->num_accel_req )* 
                                                  task_handle->task_info.pal.task_context_size,
                                                  task_handle->task_info.accel_align);
    if (task_handle->p_thread_context == NULL) {
      _ALF_DPRINTF(_ALF_ERR_LEVEL_API_, "task thread context allocate memory failed\n");
      return -ALF_ERR_NOMEM;
    }
    else _ALF_DPRINTF(_ALF_ERR_LEVEL_TRACE_, "task thread ptr allocate succeed\n");
  }
  else task_handle->task_info.pal.context_desc = NULL;

  return 0;  
}

//---------------------------------------------------------------------------------------------------
// alf_api_task_dump
//---------------------------------------------------------------------------------------------------

void alf_api_task_dump(char *prefix, alf_api_task_t * task)
{
  printf("%salf_api_task_t=%p\n", prefix, task);
  printf("%s\tself=%p\n", prefix, task->self);
  printf("%s\talf_handle=%p\n", prefix, task->alf_handle);
  printf("%s\tnext_ptr=%p\n", prefix, task->next_ptr);
  printf("%s\tattr=%u\n", prefix, task->attr);
  printf("%s\twb_dist_size=%u\n", prefix, task->wb_dist_size);
  printf("%s\tnum_accel_req=%u\n", prefix, task->num_accel_req);
  printf("%s\tstate=%d\n", prefix, task->state);
  printf("%s\tfinalized_flag=%d\n", prefix, task->finalized_flag);
  printf("%s\tstatus=%u\n", prefix, task->status);
  printf("%s\task_wbq=%p\n", prefix, task->task_wbq);
  printf("%s\taccel_wbq=%p\n", prefix, task->accel_wbq);
  printf("%s\taccel_fin_wbq=%p\n", prefix, task->accel_fin_wbq);
  printf("%s\tparent_count=%u\n", prefix, task->parent_count);
  printf("%s\tchild_tasks=%p\n", prefix, task->child_tasks);
  printf("%s\tdist_num=%u\n", prefix, task->dist_num);
  printf("%s\tcur_accel=%d\n", prefix, task->cur_accel);
  printf("%s\tp_task_threads=%p\n", prefix, task->p_task_threads);
  printf("%s\tnum_accels=%d\n", prefix, task->num_accels);
  printf("%s\tnum_wb_pending=%d\n", prefix, task->num_wb_pending);
  printf("%s\tnum_wb=%d\n", prefix, task->num_wb);
  printf("%s\tdataset=%p\n", prefix, task->dataset);
}

//---------------------------------------------------------------------------------------------------
// alf_api_wb_dump
//---------------------------------------------------------------------------------------------------

void alf_api_wb_dump(char *prefix, alf_api_wb_t * wb)
{
	char prefix_buffer[strlen(prefix)+2];
	
  printf("%salf_api_wb_t=%p\n", prefix, wb);
  
  if (wb == NULL) {
  	return;
  }
  
  printf("%s\ttask=%p\n", prefix, wb->task);
  printf("%s\tcur_dtl=%u\n", prefix, wb->cur_dtl);
  printf("%s\tcur_dtl_buffer_type=%d\n", prefix, wb->cur_dtl_buffer_type);
  
  strcpy(prefix_buffer,prefix);
  strcat(prefix_buffer,"\t");
  
  alf_pal_wb_dump(prefix_buffer,&wb->pal);
}

//---------------------------------------------------------------------------------------------------
// alf_api_dataset_dump
//---------------------------------------------------------------------------------------------------

void alf_api_dataset_dump(char *prefix, alf_api_dataset_t * dataset)
{
  unsigned int i;

  /* Lock the data set before updating and accessing it */
  pthread_mutex_lock(&dataset->lock);

  printf("%salf_api_dataset_t=%p\n", prefix, dataset);
  printf("%s\tapi_handle=%p\n", prefix, dataset->api_handle);
  printf("%s\tstate=%d\n", prefix, dataset->state);

  /* For each buffer */
  printf("%s\tbuffers=%p\n", prefix, dataset->buffers);
  for (i = 0; i < alf_arraylist_get_length(dataset->buffers); i++) {

    /* Get data set buffer address */
    alf_api_dataset_buffer_t *dataset_buffer =
        (alf_api_dataset_buffer_t *) alf_arraylist_get_element(dataset->buffers, i);

    printf("%s\tdataset_buffer=%p: addr=0x%016llx, size=0x%016llx, access_mode=%d\n", prefix, dataset_buffer, dataset_buffer->addr, dataset_buffer->size, dataset_buffer->access_mode);

  }

  /* For each task */
  printf("%s\ttasks=%p\n", prefix, dataset->tasks);
  for (i = 0; i < alf_arraylist_get_length(dataset->tasks); i++) {

    /* Get task address */
    alf_api_task_t *dataset_task = (alf_api_task_t *) alf_arraylist_get_element(dataset->tasks, i);

    printf("%s\tdataset_task=%p\n", prefix, dataset_task);

  }

  printf("%s\tpal_dataset=%p\n", prefix, dataset->pal_dataset);

  if (dataset->pal_dataset != NULL) {
  	char prefix_buffer[strlen(prefix)+2];

    strcpy(prefix_buffer,prefix);
    strcat(prefix_buffer,"\t");
  
    alf_pal_dataset_dump(prefix_buffer, dataset->pal_dataset);
  }

  /* Unlock data set object */
  pthread_mutex_unlock(&dataset->lock);

}
