/* --------------------------------------------------------------  */
/* (C)Copyright 2007                                               */
/* International Business Machines Corporation                     */
/* All Rights Reserved.                                            */
/*                                                                 */
/* Redistribution and use in source and binary forms, with or      */
/* without modification, are permitted provided that the           */
/* following conditions are met:                                   */
/*                                                                 */
/* Redistributions of source code must retain the above copyright  */
/* notice, this list of conditions and the following disclaimer.   */
/*                                                                 */
/* Redistributions in binary form must reproduce the above         */
/* copyright notice, this list of conditions and the following     */
/* disclaimer in the documentation and/or other materials          */
/* provided with the distribution.                                 */
/*                                                                 */
/* Neither the name of IBM Corporation nor the names of its        */
/* contributors may be used to endorse or promote products         */
/* derived from this software without specific prior written       */
/* permission.                                                     */
/*                                                                 */
/* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND          */
/* CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,     */
/* INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF        */
/* MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE        */
/* DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR            */
/* CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,    */
/* SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT    */
/* NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;    */
/* LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)        */
/* HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN       */
/* CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR    */
/* OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,  */
/* EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.              */
/* --------------------------------------------------------------  */

/*
 ************************************************************
 * ALF Host Scheduler
 ************************************************************
 */

#include <stddef.h>
#include <string.h>
#include <stdio.h>

//#define _ALF_DEBUG_LEVEL_           10   //  0 to 10,
#include <alf.h>

#include "alf_host.h"
#include "alf_debug.h"
#include "alf_api_local.h"
#include "alf_trace.h"
#include "alf_hooks_perform_host.h"

/*
 ************************************************************
 * Globals
 ************************************************************
 */



/*
 ************************************************************
 * Local Globals
 ************************************************************
 */


/*
 ************************************************************
 * Local APIs
 ************************************************************
 */

/* This is the prototype of ALF scheduler */

void *alf_api_scheduler(void *alf_ptr)
{
  alf_api_t *alf_handle = (alf_api_t *) alf_ptr;
  _ALF_DPRINTF(_ALF_ERR_LEVEL_ENTRYEXIT_, "Enter: alf_ptr=%p\n", alf_ptr);
  do {
    if (alf_handle->state == ALF_API_STATUS_EXITING) {
      _ALF_DPRINTF(_ALF_ERR_LEVEL_TRACE_, "alf_handle->state=ALF_API_STATUS_EXITING\n");
      if (alf_handle->exit_policy == ALF_EXIT_POLICY_FORCE) {
        //force exit, cancel all tasks in init/ready/exec list
        alf_sched_task_cancel(alf_handle);
        break;
      }
      if (alf_arraylist_is_empty(alf_handle->init_task_list)
          && alf_arraylist_is_empty(alf_handle->ready_task_list)
          && alf_arraylist_is_empty(alf_handle->exec_task_list)) {
        //non-force exit, only exit when there's no running or pending task
        break;
      }
    }

    if (alf_arraylist_is_empty(alf_handle->ready_task_list)
        || alf_thread_mgr_query(alf_handle)) {
      //if ready list is empty or threads still available, call task select
      alf_sched_task_select(alf_handle);
    }
    
    //pick up task to run
    alf_sched_task_schedule(alf_handle);

    //do work block scheduling for each running task
    alf_sched_task_run(alf_handle);

    //release resources of finished or destroyed tasks
    alf_sched_task_release(alf_handle);
  }
  while (1);

  _ALF_DPRINTF(_ALF_ERR_LEVEL_TRACE_, "Signal alf_handle->cond with alf_handle->state = ALF_API_STATUS_EXITED\n");
  _ALF_API_STD_COND_SIG(alf_handle->lock, alf_handle->cond, alf_handle->state = ALF_API_STATUS_EXITED);

	_ALF_DPRINTF(_ALF_ERR_LEVEL_ENTRYEXIT_, "Exit: ret=%p\n", NULL);
  return NULL;
}

void alf_sched_pal_wb_enqueue(alf_api_task_t* task_handle, int i, alf_api_wb_t *wb)
{
  int pal_ret = 0;

	_ALF_DPRINTF(_ALF_ERR_LEVEL_ENTRYEXIT_, "Enter: task_handle=%p, i=%d, wb=%p\n", task_handle, i, wb);
   pal_ret = alf_pal_wb_enqueue(task_handle->p_task_threads[i].thread, 
                                (alf_pal_wb_handle) wb);
   if( 0 == pal_ret )
   {
     //enqueue to garbage queue failed will cause memory leak
     //but we can still go on
     if( NULL == alf_arraylist_enqueue(task_handle->accel_fin_wbq[i], 
                                        (alf_pal_wb_handle) wb))
     {
       alf_err_error_processing(task_handle->alf_handle, ALF_ERR_WARNING, ALF_ERR_NOMEM, 
                                  "WB:enqueue to garbage queue failed\n");
     }
   }
   //if pal enqueue failed, call error handler
   else
   {
     alf_err_error_processing(task_handle->alf_handle, ALF_ERR_WARNING, pal_ret, 
                                  "WB:enqueue to PAL queue failed\n");
   }
	_ALF_DPRINTF(_ALF_ERR_LEVEL_ENTRYEXIT_, "Exit\n");
}

/* cyclic distribution
 * dispatch wb from accel wb queue to pal wb queue directly
 */
int alf_accel_wb_cyclic_dispatch(alf_api_task_t * task_handle)
{
  alf_api_wb_t *wb;
  int i;
  int ret = 0;
	_ALF_DPRINTF(_ALF_ERR_LEVEL_ENTRYEXIT_, "Enter: task_handle=%p\n", task_handle);
  for (i = 0; i < (int) task_handle->num_accels; i++) {
    //dispatch as much as possible wb to corresponding thread
    while (!alf_arraylist_is_empty(task_handle->accel_wbq[i])) {
      if (alf_pal_wbqueue_query(task_handle->p_task_threads[i].thread, NULL) > 0) {
        wb = alf_arraylist_dequeue(task_handle->accel_wbq[i]);
        alf_sched_pal_wb_enqueue(task_handle, i, wb);
      } else {
        ret = 1;
        break;
      }
    }
  }
	_ALF_DPRINTF(_ALF_ERR_LEVEL_ENTRYEXIT_, "Exit: ret=%d\n", ret);
  return ret;
}



/* non-cyclic distribution
 */
int alf_accel_wb_noncyclic_dispatch(alf_api_task_t * task_handle)
{
  alf_api_wb_t *wb;
  unsigned int accel_index;
  unsigned int wb_index;
  int accel_dist_num;
  int ret = 0;
	_ALF_DPRINTF(_ALF_ERR_LEVEL_ENTRYEXIT_, "Enter: task_handle=%p\n", task_handle);
  //if no wb bundle, no accel wb queue needed
  //dispatch from task wb queue to pal wb queue
  if (task_handle->wb_dist_size == 1) {
    accel_index = 0;
    while (!alf_arraylist_is_empty(task_handle->task_wbq)) {
      if (accel_index >= task_handle->num_accels) {
        //all accel pal queue is full
        ret = 1;
        break;
      }
      //dispatch wb as much as possible to pal queue
      if (alf_pal_wbqueue_query(task_handle->p_task_threads[accel_index].thread, NULL) > 0) {
        wb = alf_arraylist_dequeue(task_handle->task_wbq);
        alf_sched_pal_wb_enqueue(task_handle, accel_index, wb);
      } else {
        //current accel is full, turn to next accel
        accel_index++;
      }
    }                           //while  
  }                             //if
  else {
    //if bundled, first try to dispatch wb in accel wb queue to pal wb queue
    //if accel wb queue is empty, try to dispatch as much as possible wb to pal wb queue
    //if pal wb queue is full and one bundle not finished,buffer left wb in accel wb queue
    for (accel_index = 0; accel_index < task_handle->num_accels; accel_index++) {
      accel_dist_num = 0;
      while (!alf_arraylist_is_empty(task_handle->accel_wbq[accel_index])) {
        //dispatch wb in accel wb queue to pal wb queue 
        if (alf_pal_wbqueue_query(task_handle->p_task_threads[accel_index].thread, NULL) > 0) {
          wb = alf_arraylist_dequeue(task_handle->accel_wbq[accel_index]);
          alf_sched_pal_wb_enqueue(task_handle, accel_index, wb);
          _ALF_DPRINTF_X(_ALF_ERR_LEVEL_TRACE_, "accel %d dispatch from buffer queue to pal queue\n", accel_index);
        } else {
          ret = 1;
          break;
        }
      }
      if (alf_arraylist_is_empty(task_handle->accel_wbq[accel_index])) {
        //dispatch wb to pal wb queue as much as possible
        //skip if less than one bundle wb in task wb queue
        //last incomplete bundle should be processed after this function return success 
        while (alf_arraylist_get_length(task_handle->task_wbq) >= task_handle->wb_dist_size) {
            for (wb_index = 0; wb_index < task_handle->wb_dist_size; wb_index++) {
              if (alf_pal_wbqueue_query(task_handle->p_task_threads[accel_index].thread, NULL) > 0)
              {
                wb = alf_arraylist_dequeue(task_handle->task_wbq);
                alf_sched_pal_wb_enqueue(task_handle, accel_index, wb);
                _ALF_DPRINTF_X(_ALF_ERR_LEVEL_TRACE_, "accel %d dispatch from task queue to pal queue\n", accel_index);
              } else
                break;
            }
            //one bundle not finished, buffer left wb in pal wb queue
            if (wb_index < task_handle->wb_dist_size) {
              while (wb_index < task_handle->wb_dist_size) {
                wb = alf_arraylist_dequeue(task_handle->task_wbq);
                if(NULL == alf_arraylist_enqueue(task_handle->accel_wbq[accel_index], 
                                                    (alf_pal_wb_handle) wb))
                {
                   alf_err_error_processing(task_handle->alf_handle, ALF_ERR_WARNING, 
                                       ALF_ERR_NOMEM, 
                                       "WB:enqueue to garbage queue failed\n");
                }
                wb_index++;
                _ALF_DPRINTF_X(_ALF_ERR_LEVEL_TRACE_, "accel %d put into buffer queue\n", accel_index);
              }
              //turn to next accel
              ret = 1;
              break;
            }
        }                       //while alf_arraylist_get_length
      }                         //if 
    }                           //for
  }                             //else
	_ALF_DPRINTF(_ALF_ERR_LEVEL_ENTRYEXIT_, "Exit: ret=%d\n", ret);
  return ret;
}
/*
 * dispatch last incomplete wb bundle for noncyclic bundled task
 *
 */
int alf_accel_wb_noncyclic_incomp_bundle_dispatch(alf_api_task_t * task_handle)
{
  alf_api_wb_t *wb;
  
  if (task_handle->attr & ALF_TASK_ATTR_WB_CYCLIC || task_handle->wb_dist_size <= 1) {
    return 0;
  }
  else
  {
        //dispatch all wb in last incomplete bundle to accel 0
        while (!alf_arraylist_is_empty(task_handle->task_wbq)) {
              if (alf_pal_wbqueue_query(task_handle->p_task_threads[0].thread, NULL) > 0)
              {
                wb = alf_arraylist_dequeue(task_handle->task_wbq);
                alf_sched_pal_wb_enqueue(task_handle, 0 , wb);
                _ALF_DPRINTF_X(_ALF_ERR_LEVEL_TRACE_, "accel 0 dispatch from task queue to pal queue\n");
              } 
              else
              {
                return 1; 
              }
        }//while
        return 0;
  }
}
int alf_sched_task_wb_dispatch(alf_api_task_t * task_handle)
{
	int ret;
	_ALF_DPRINTF(_ALF_ERR_LEVEL_ENTRYEXIT_, "Enter: task_handle=%p\n", task_handle);
  if (task_handle->attr & ALF_TASK_ATTR_WB_CYCLIC) {
    ret = alf_accel_wb_cyclic_dispatch(task_handle);
  } else {
    ret = alf_accel_wb_noncyclic_dispatch(task_handle);
  }
	_ALF_DPRINTF(_ALF_ERR_LEVEL_ENTRYEXIT_, "Exit: ret=%d\n", ret);
	return ret;
}



/*
*  alf task selection 
*  pick up ready task from init list and put into ready list
*/
void alf_sched_task_select(alf_api_t * alf_handle)
{
  unsigned int i;
  alf_api_task_t *task_handle;
	_ALF_DPRINTF(_ALF_ERR_LEVEL_ENTRYEXIT_, "Enter: alf_handle=%p\n", alf_handle);
  unsigned int task_num = alf_arraylist_get_length(alf_handle->init_task_list);
  for (i = 0; i < task_num; i++) {
    task_handle = (alf_api_task_t *) alf_arraylist_dequeue(alf_handle->init_task_list);
    pthread_mutex_lock(&task_handle->lock);
    //task not ready, put it back into init list
    if (task_handle->state == ALF_API_TASK_STATUS_INIT)
    {
      //finalized task with no wb enqueued should also be selected to run 
      if(task_handle->finalized_flag && task_handle->parent_count == 0)
      {
        task_handle->state = ALF_API_TASK_STATUS_READY;
        pthread_mutex_unlock(&task_handle->lock);
        if(NULL == alf_arraylist_enqueue(alf_handle->ready_task_list, task_handle))
        {
          alf_err_error_processing(task_handle->alf_handle, ALF_ERR_EXCEPTION, 
                                       ALF_ERR_NOMEM, 
                                       "TASK:enqueue to ready list failed\n");
        }
      }
      else
      {
        pthread_mutex_unlock(&task_handle->lock);
        if(NULL == alf_arraylist_enqueue(alf_handle->init_task_list, task_handle))
        {
          alf_err_error_processing(task_handle->alf_handle, ALF_ERR_EXCEPTION, 
                                       ALF_ERR_NOMEM, 
                                       "TASK:enqueue to init list failed\n");
        }
      }
    }
    //if task ready to run, put into ready list
    else if (task_handle->state == ALF_API_TASK_STATUS_PENDING )
    {
      if( task_handle->parent_count == 0 )
      {
        task_handle->state = ALF_API_TASK_STATUS_READY;
        pthread_mutex_unlock(&task_handle->lock);
        if(NULL == alf_arraylist_enqueue(alf_handle->ready_task_list, task_handle))
        {
          alf_err_error_processing(task_handle->alf_handle, ALF_ERR_EXCEPTION, 
                                       ALF_ERR_NOMEM, 
                                       "TASK:enqueue to ready list failed\n");
        }
      }
      else
      {
        pthread_mutex_unlock(&task_handle->lock);
        if(NULL == alf_arraylist_enqueue(alf_handle->init_task_list, task_handle))
        {
          alf_err_error_processing(task_handle->alf_handle, ALF_ERR_EXCEPTION, 
                                       ALF_ERR_NOMEM, 
                                       "TASK:enqueue to init list failed\n");
        }
      }  
    } 
    //task destroyed
    else
    {
      pthread_mutex_unlock(&task_handle->lock);
      if(NULL == alf_arraylist_enqueue(alf_handle->destroy_task_list, task_handle))
      {
        //only fail when list is full and no memory for realloc
        //release task resource here
        alf_int_task_res_destroy(task_handle);
      }
    }
  }
	_ALF_DPRINTF(_ALF_ERR_LEVEL_ENTRYEXIT_, "Exit\n");
}



/*
 * alf task scheduling
 * schedule task to run on pal thread
*/
void alf_sched_task_schedule(alf_api_t * alf_handle)
{
  unsigned int i;
  alf_api_task_t *task_handle;
	_ALF_DPRINTF(_ALF_ERR_LEVEL_ENTRYEXIT_, "Enter: alf_handle=%p\n", alf_handle);
  unsigned int task_num = alf_arraylist_get_length(alf_handle->ready_task_list);
  int rtn;

  for (i = 0; i < task_num; i++) {
    //if no thread left, skip the left tasks
    if(alf_thread_mgr_query(alf_handle) <= 0)
    {
      break;
    }
    task_handle = (alf_api_task_t *) alf_arraylist_dequeue(alf_handle->ready_task_list);
    pthread_mutex_lock(&task_handle->lock);
    if (task_handle->state == ALF_API_TASK_STATUS_READY) {
      if(task_handle->p_context != NULL 
        && task_handle->task_info.ctx_entry_size != 0 )
      {
        memcpy(task_handle->p_thread_context, task_handle->p_context, 
           task_handle->task_info.ctx_entry_size);
      }
      //if threads allocated successfully, put into exec list to run
      rtn = alf_thread_mgr_start_thread(alf_handle, task_handle);
      if( rtn > 0 )
      {
        //create accel wbq and garbage queue for non-fixed mapping
        if (alf_sched_task_wbq_create(task_handle) < 0)
        {
          //if failed, return thread and put back to ready list
          alf_thread_mgr_return_thread(alf_handle, task_handle);
          pthread_mutex_unlock(&task_handle->lock);
          alf_arraylist_enqueue(alf_handle->ready_task_list, task_handle);
        }
        if (task_handle->state == ALF_API_TASK_STATUS_READY) {

          //alf performance hook end ALF_TASK_BEFORE_EXEC_INTERVAL
	  TRACE_INTERVAL_END(_ALF_TASK_BEFORE_EXEC_INTERVAL,task_handle->task_before_exec_token,1,task_handle);
          task_handle->state = ALF_API_TASK_STATUS_EXEC;
          //alf performance hook begin ALF_TASK_EXEC_INTERVAL
          TRACE_INTERVAL_BEGIN(_ALF_TASK_EXEC_INTERVAL,task_handle->task_in_exec_token,1);

        }
        pthread_mutex_unlock(&task_handle->lock);
        if(NULL == alf_arraylist_enqueue(alf_handle->exec_task_list, task_handle))
        {
          alf_err_error_processing(task_handle->alf_handle, ALF_ERR_EXCEPTION, 
                                       ALF_ERR_NOMEM, 
                                       "TASK:enqueue to exec list failed\n");
        }
        alf_int_task_call_event_handler(task_handle, ALF_TASK_EVENT_READY);
      } else {
        //required threads not available, so put back into ready list
        if( rtn == -ALF_ERR_AGAIN)
        {
          pthread_mutex_unlock(&task_handle->lock);
          //should not fail here, since just put back it will not cause queue size exceeding
          alf_arraylist_enqueue(alf_handle->ready_task_list, task_handle);
        }
        else
        {
          // Thread creation failed fatally.  Mark thread as dead, and
          // broadcast out a signal to unblock any waiting user threads
	  // The task mutex is already (somewhat dubiously) held
          task_handle->state = ALF_API_TASK_STATUS_DESTROY;
          pthread_cond_signal(&task_handle->cond);
	  
          pthread_mutex_unlock(&task_handle->lock);
          alf_err_error_processing(task_handle->alf_handle, ALF_ERR_EXCEPTION, 
                                       rtn, 
                                       "TASK:start pal thread failed\n");
          
        }
      }
    } else
      //task destroyed
    {
      pthread_mutex_unlock(&task_handle->lock);
      if(NULL == alf_arraylist_enqueue(alf_handle->destroy_task_list, task_handle))
      {
        //only fail when list is full and no memory for realloc
        //release task resource here
        alf_int_task_res_destroy(task_handle);
      }
    }
  }
	_ALF_DPRINTF(_ALF_ERR_LEVEL_ENTRYEXIT_, "Exit\n");
}



/* alf task stop
 * do the following things to stop a task:
 * free all wb
 * context merge
 * call event handler
 * send finish signal
 * update parent count
 * release threads
*/
int alf_task_stop(alf_api_t * alf_handle, alf_api_task_t * task_handle)
{
	_ALF_DPRINTF(_ALF_ERR_LEVEL_ENTRYEXIT_, "Enter: alf_handle=%p, task_handle=%p\n", alf_handle, task_handle);
  //free left wb
  alf_sched_task_wb_free(task_handle);

  //context merge
  if(task_handle->p_context != NULL 
     && task_handle->task_info.ctx_entry_size != 0 )
  {
    alf_sched_task_context_merge(task_handle,0);
  }

  //release threads
  alf_thread_mgr_return_thread(alf_handle, task_handle);

  //call event handler
  alf_int_task_call_event_handler(task_handle, ALF_TASK_EVENT_FINISHED);
  
  //send finish signal
  _ALF_API_STD_COND_SIG(task_handle->lock, task_handle->cond, task_handle->state = ALF_API_TASK_STATUS_FINISH);
 
  //alf perfomance hook end ALF_TASK_EXEC_INTERVAL
  TRACE_INTERVAL_END(_ALF_TASK_EXEC_INTERVAL,task_handle->task_in_exec_token,1,task_handle);
 
  //update child tasks parent count
  alf_sched_task_child_notify(task_handle);
  
	_ALF_DPRINTF(_ALF_ERR_LEVEL_ENTRYEXIT_, "Exit: ret=0\n");
  return 0;
}


int alf_task_threads_wait(alf_api_task_t *task_handle)
{
  unsigned int i;
	_ALF_DPRINTF(_ALF_ERR_LEVEL_ENTRYEXIT_, "Enter: task_handle=%p\n", task_handle);
  //wait for all thread idle
  for (i = 0; i < task_handle->num_accels; i++) {
    if (0 != alf_pal_thread_wait(task_handle->p_task_threads[i].thread, 0)) {
      _ALF_DPRINTF(_ALF_ERR_LEVEL_ENTRYEXIT_, "Exit: ret=1\n");
      return 1;
    }
  }
  _ALF_DPRINTF(_ALF_ERR_LEVEL_ENTRYEXIT_, "Exit: ret=0\n");
  return 0;
}
/* alf task run 
 * do work block scheduling for a task
*/
void alf_sched_task_run(alf_api_t * alf_handle)
{
  unsigned int i;
  alf_api_task_t *task_handle;
	_ALF_DPRINTF(_ALF_ERR_LEVEL_ENTRYEXIT_, "Enter: alf_handle=%p\n", alf_handle);
  unsigned int task_num = alf_arraylist_get_length(alf_handle->exec_task_list);
  for (i = 0; i < task_num; i++) {
    task_handle = (alf_api_task_t *) alf_arraylist_dequeue(alf_handle->exec_task_list);
    if (task_handle->state == ALF_API_TASK_STATUS_EXEC) {
      //task finalized, try to stop it
      if (task_handle->finalized_flag) {
        if (alf_sched_task_wb_dispatch(task_handle)) {
          //try to free some memory
          alf_sched_task_wb_try_free(task_handle);
          //not all wb dispatched, put back to exec list
          alf_arraylist_enqueue(alf_handle->exec_task_list, task_handle);
          continue;
        }
        if(alf_accel_wb_noncyclic_incomp_bundle_dispatch(task_handle))
        {
	  // No free slots in accel queue, try to free completed blocks
          alf_sched_task_wb_try_free(task_handle);

          //non-cyclic last incomplete bundle not finished, so put back on queue
          alf_arraylist_enqueue(alf_handle->exec_task_list, task_handle);

          continue;
        }
        if(alf_task_threads_wait(task_handle))
        {
          //thread is still busy, put back to exec list
          alf_arraylist_enqueue(alf_handle->exec_task_list, task_handle);
          continue;
        }
        alf_task_stop(alf_handle,task_handle);
        if(NULL == alf_arraylist_enqueue(alf_handle->destroy_task_list, task_handle))
        {
          //should not fail here, since just put back it will not cause queue size exceeding
          alf_arraylist_enqueue(alf_handle->exec_task_list, task_handle);
        }
      //task still running
      } else {
        alf_sched_task_wb_dispatch(task_handle);
        //try to free some memory
        alf_sched_task_wb_try_free(task_handle);
        //should not fail here, since just put back it will not cause queue size exceeding
        alf_arraylist_enqueue(alf_handle->exec_task_list, task_handle);
      }
    }                           //if
    else if (task_handle->state == ALF_API_TASK_STATUS_DESTROY) {
      // the following will stop the threads if they are still busy
      alf_thread_mgr_return_thread(alf_handle, task_handle);
      if(NULL == alf_arraylist_enqueue(alf_handle->destroy_task_list, task_handle))
      {
        //only fail when list is full and no memory for realloc
        //release task resource here
        alf_int_task_res_destroy(task_handle);
      }
    } //else
  }
  _ALF_DPRINTF(_ALF_ERR_LEVEL_ENTRYEXIT_, "Exit\n");
}

/* update parent count of it's child tasks when finished
**/
void alf_sched_task_child_notify(alf_api_task_t * task_handle)
{
  int child_num;
  int i;

	_ALF_DPRINTF(_ALF_ERR_LEVEL_ENTRYEXIT_, "Enter: task_handle=%p\n", task_handle);
  alf_api_task_t *child_task;
  child_num = alf_arraylist_get_length(task_handle->child_tasks);
  for (i = 0; i < child_num; i++) {
    child_task = alf_arraylist_dequeue(task_handle->child_tasks);
    pthread_mutex_lock(&child_task->lock);
    child_task->parent_count--;
    pthread_mutex_unlock(&child_task->lock);
  }
  _ALF_DPRINTF(_ALF_ERR_LEVEL_ENTRYEXIT_, "Exit\n");
}


/* merge context in parallel
 * merge neibourhood context in pair and recursively
 *  0<-1 2<-3 4<-5 6<-7 
 *  0<-2      4<-6
 *  0<-4
**/
void alf_sched_task_context_merge(alf_api_task_t* task_handle, unsigned int depth)
{
	_ALF_DPRINTF(_ALF_ERR_LEVEL_ENTRYEXIT_, "Enter: task_handle=%p, depth=%d\n", task_handle, depth);
  unsigned int i;
  unsigned int ctx_size = task_handle->task_info.pal.task_context_size;
  int jump = 1<<(depth+1);
  int next = 1<<depth;
  alf_data_addr64_t ctx_addr;
  int ret;
  if( (task_handle->num_accels == 1 && depth == 1)
      || (depth > 1 && (1<<(depth-1)) >= (int)task_handle->num_accels) )
  {
      while(1)
      {
        ret = alf_pal_thread_wait(task_handle->p_task_threads[0].thread, 0);
        if( 0 == ret ) break;
        else if( ret < 0) return;
      } 
     _ALF_DPRINTF_X(_ALF_ERR_LEVEL_TRACE_,"wait thread %d\n",0);
      PTR_TO_ADDR64(task_handle->p_thread_context,ctx_addr);
      alf_pal_thread_context_swap(task_handle->p_task_threads[0].thread,0,
                                  ctx_addr);  
      _ALF_DPRINTF_X(_ALF_ERR_LEVEL_TRACE_,"swap thread %d to context %d\n",0,0);
      while(1)
      {
        ret = alf_pal_thread_wait(task_handle->p_task_threads[0].thread, 0);
        if( 0 == ret ) break;
        else if( ret < 0) return;
      } 
      memcpy(task_handle->p_context, task_handle->p_thread_context, 
             task_handle->task_info.ctx_entry_size);
      _ALF_DPRINTF(_ALF_ERR_LEVEL_ENTRYEXIT_, "Exit\n");
      return;
  }
  //swap out odd thread
  //1 3 5 7, start = 1, jump = 2 
  //2 6, start = 2, jump = 4
  //4, start = 4, jump = 8
  for(i=next;i<task_handle->num_accels;i+=jump)
  {
      while(1)
      {
        ret = alf_pal_thread_wait(task_handle->p_task_threads[i].thread, 0);
        if( 0 == ret ) break;
        else if( ret < 0) return;
      } 
      _ALF_DPRINTF_X(_ALF_ERR_LEVEL_TRACE_,"wait thread %d\n",i);
      PTR_TO_ADDR64(&((char *)task_handle->p_thread_context)[ctx_size*i],ctx_addr);
      alf_pal_thread_context_swap(task_handle->p_task_threads[i].thread,0,
                                  ctx_addr);  
      _ALF_DPRINTF_X(_ALF_ERR_LEVEL_TRACE_,"swap thread %d to context %d\n",i,i);
  }
  //merge even thread with odd thread
  //0<-1 2<-3 4<-5 6<-7,jump = 2
  //0<-2 4<-6, jump = 4
  //0<-4, jump = 8 
  for(i=0;i<task_handle->num_accels;i+=jump)
  {
    if(i+next<task_handle->num_accels)
    {
      while(1)
      {
        ret = alf_pal_thread_wait(task_handle->p_task_threads[i+next].thread, 0);
        if( 0 == ret ) break;
        else if( ret < 0) return;
      } 
        _ALF_DPRINTF_X(_ALF_ERR_LEVEL_TRACE_,"wait thread %d\n",i+next);
        PTR_TO_ADDR64(&((char *)task_handle->p_thread_context)[ctx_size*(i+next)],ctx_addr);
        alf_pal_thread_context_merge(task_handle->p_task_threads[i].thread,ctx_addr);  
        _ALF_DPRINTF_X(_ALF_ERR_LEVEL_TRACE_,"merge context %d to thread %d\n",i+next,i);
    }
  }
  alf_sched_task_context_merge(task_handle,depth+1);
  _ALF_DPRINTF(_ALF_ERR_LEVEL_ENTRYEXIT_, "Exit\n");
}
