/* --------------------------------------------------------------  */
/* (C)Copyright 2007                                               */
/* International Business Machines Corporation                     */
/* All Rights Reserved.                                            */
/*                                                                 */
/* Redistribution and use in source and binary forms, with or      */
/* without modification, are permitted provided that the           */
/* following conditions are met:                                   */
/*                                                                 */
/* Redistributions of source code must retain the above copyright  */
/* notice, this list of conditions and the following disclaimer.   */
/*                                                                 */
/* Redistributions in binary form must reproduce the above         */
/* copyright notice, this list of conditions and the following     */
/* disclaimer in the documentation and/or other materials          */
/* provided with the distribution.                                 */
/*                                                                 */
/* Neither the name of IBM Corporation nor the names of its        */
/* contributors may be used to endorse or promote products         */
/* derived from this software without specific prior written       */
/* permission.                                                     */
/*                                                                 */
/* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND          */
/* CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,     */
/* INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF        */
/* MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE        */
/* DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR            */
/* CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,    */
/* SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT    */
/* NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;    */
/* LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)        */
/* HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN       */
/* CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR    */
/* OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,  */
/* EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.              */
/* --------------------------------------------------------------  */

#include <stdlib.h>
#include <string.h>
#include "alf_host.h"
#include "alf_debug.h"
#include "alf_hooks_perform_host.h"

// TODO: need some smarter algorithm for the detection of image change

// create a copy of the image info

static int alf_thread_mgr_image_info_copy(alf_thread_pool_elem_t *elem,  alf_api_task_info_t *p_info)
{
    strncpy(elem->lib_name, p_info->pal.api_str[ALF_API_KERNEL_LIBRARY], ALF_STRING_TOKEN_MAX);
    elem->lib_name[ALF_STRING_TOKEN_MAX] = 0;
    strncpy(elem->image_name, p_info->pal.api_str[ALF_API_KERNEL_IMAGE], ALF_STRING_TOKEN_MAX);
    elem->image_name[ALF_STRING_TOKEN_MAX] = 0;
    return 0;
}


// return 0 when the info is identical

static int alf_thread_mgr_image_info_cmp(alf_thread_pool_elem_t *elem,  alf_api_task_info_t *p_info)
{
	if(strcmp(elem->lib_name, p_info->pal.api_str[ALF_API_KERNEL_LIBRARY]) == 0 &&
	   strcmp(elem->image_name, p_info->pal.api_str[ALF_API_KERNEL_IMAGE]) == 0 )
	  return 1;
        else return 0;
}


/*
 * Setup the thread manager
 * will allocate the thread pool to cache
 * the PAL thread used before
 */
int 
alf_thread_mgr_setup(alf_api_t* alf_handle, unsigned int num_of_accels)
{
    unsigned int i;
    int rtn;
    
    pthread_mutex_init(&alf_handle->thread_lock, NULL);
    
    // check if we have already did the init
    if (alf_handle->accel_num > 0)	// we already did the init
  	{
  		return -ALF_ERR_PERM;	// not permitted
  	}


    //  try to reserve from PAL layer
    rtn = alf_pal_accelerators_reserve (alf_handle->platform_handle,
				 num_of_accels,
				 ALF_ACCEL_RESV_POLICY_PERSIST, -1);
    if(rtn < 0)
  	{
	  return -ALF_ERR_PERM;   // not permitted
  	}

	num_of_accels = (unsigned)rtn;  // this is what we actually get

    // get the pool elements
	alf_handle->threadpool.elems = (alf_thread_pool_elem_t *)
                         calloc(num_of_accels, sizeof(alf_thread_pool_elem_t));
	if(alf_handle->threadpool.elems == NULL)
		{
			return -ALF_ERR_NOMEM;  // or -ALF_ERR_GENERIC ?
		}
	

	alf_handle->accel_num = num_of_accels;
    alf_handle->threadpool.num_threads = alf_handle->accel_num;

    alf_handle->threadpool.used_threads = 0;  // not yet used

    // initial the states
	for(i=0; i<num_of_accels; i++) 
    {
       alf_handle->threadpool.elems[i].status = ALF_API_THREAD_STATUS_INIT;
    }
 
    return 0;   
}



/*
 * Start all the PAL threads for a task
 * If there are some PAL threads cache, reuse them
 */
int
alf_thread_mgr_start_thread(alf_api_t* alf_handle, alf_api_task_t* task_handle)
{
  unsigned int allocated_accels = 0;
  unsigned int i;
  unsigned int num_avail = alf_handle->threadpool.num_threads - alf_handle->threadpool.used_threads;
  int rtn;
  
  if(num_avail == 0 
  	|| (task_handle->num_accel_req > num_avail && (task_handle->attr & ALF_TASK_ATTR_SCHED_FIXED)) ) 
  {
  	return -ALF_ERR_AGAIN;	// we have no resource
  }
  // we do not deal with increamental now, maybe in the future
  if( task_handle->num_accels > 0 )
  {
    return -ALF_ERR_BADR;
  }

  // first, check for cached threads to see if we can reuse it

  // TODO: need more advanced methods to check for image changes
  pthread_mutex_lock(&alf_handle->thread_lock);
  for(i=0; i<alf_handle->threadpool.num_threads; i++) 
  {
    if((alf_handle->threadpool.elems[i].status == ALF_API_THREAD_STATUS_STANDBY)  && 
		alf_thread_mgr_image_info_cmp(&alf_handle->threadpool.elems[i], &task_handle->task_info))
    {
      //Reset PAL Thread Here
        rtn = alf_pal_thread_reset((alf_pal_task_info_t *) & task_handle->
                             task_info, alf_handle->threadpool.elems[i].task_thread);
        if(rtn < 0)
        {
            // we need to clean up allocated threads so far
            goto label_thread_mgr_start_err;
        }
        if( task_handle->p_thread_context != 0)
        {
          //update task context after reset
          rtn = alf_pal_thread_context_swap(alf_handle->threadpool.elems[i].task_thread,
#ifdef _ALF_64B_
					  (alf_data_addr64_t)task_handle->p_thread_context,0);
#else
					  (alf_data_addr64_t)((unsigned int)task_handle->p_thread_context) , 0 );
#endif
        }
        if(rtn < 0)
        {
            // we need to clean up allocated threads so far
            goto label_thread_mgr_start_err;
        }

      alf_handle->threadpool.elems[i].status = ALF_API_THREAD_STATUS_RUNNING;
	    task_handle->p_task_threads[allocated_accels].thread = alf_handle->threadpool.elems[i].task_thread;
	    task_handle->p_task_threads[allocated_accels].finished_wbs = 0;  // just started
	    task_handle->p_task_threads[allocated_accels].accel_id = i;
      alf_handle->threadpool.used_threads ++;
      allocated_accels ++;
      //alf performance hook ALF_THREAD_RUN_INTERVAL begin
      TRACE_INTERVAL_BEGIN(_ALF_THREAD_RUN_INTERVAL,alf_handle->threadpool.elems[i].thread_run_token,1);

      alf_int_task_call_event_handler(task_handle, ALF_TASK_EVENT_INSTANCE_START);
    }
  
    if(allocated_accels == task_handle->num_accel_req) 
    {
		goto label_thread_mgr_start_done;
    }
  }


  // then, check for unused threads
  
  for(i=0; i<alf_handle->threadpool.num_threads; i++)
  {
    if(alf_handle->threadpool.elems[i].status == ALF_API_THREAD_STATUS_INIT)
    {
       rtn = alf_pal_thread_create(alf_handle->platform_handle, i, 
                              (alf_pal_task_info_t*)&task_handle->task_info, 
#ifdef _ALF_64B_
					  (alf_data_addr64_t)task_handle->p_thread_context,
#else
					  (alf_data_addr64_t)((unsigned int)task_handle->p_thread_context),
#endif
                     &task_handle->p_task_threads[allocated_accels].thread);
       if(rtn < 0)
       {
            // we need to clean up allocated threads so far
            goto label_thread_mgr_start_err;
       }

	    task_handle->p_task_threads[allocated_accels].finished_wbs = 0; // just started
	    task_handle->p_task_threads[allocated_accels].accel_id = i;

	    rtn = alf_thread_mgr_image_info_copy(&alf_handle->threadpool.elems[i], &task_handle->task_info);
      if(rtn < 0)
      {
            // we need to clean up allocated threads so far
            alf_pal_thread_destroy(task_handle->p_task_threads[allocated_accels].thread);
            goto label_thread_mgr_start_err;
      }
	  
      alf_handle->threadpool.elems[i].task_thread = 
                               task_handle->p_task_threads[allocated_accels].thread;
      alf_handle->threadpool.elems[i].status = ALF_API_THREAD_STATUS_RUNNING;
      alf_handle->threadpool.used_threads ++;
      allocated_accels ++;

      //alf performance hook ALF_THREAD_RUN_INTERVAL begin
      TRACE_INTERVAL_BEGIN(_ALF_THREAD_RUN_INTERVAL,alf_handle->threadpool.elems[i].thread_run_token,1);

      alf_int_task_call_event_handler(task_handle, ALF_TASK_EVENT_INSTANCE_START);
    }

    if(allocated_accels == task_handle->num_accel_req) 
    {
		goto label_thread_mgr_start_done;
    }
  }


  // lastly, go for more expensive cache flush

  for(i=0; i<alf_handle->threadpool.num_threads; i++)
  {
    if(alf_handle->threadpool.elems[i].status == ALF_API_THREAD_STATUS_STANDBY)
    {
       //destroy the thread first, then alloc 
       rtn = alf_pal_thread_destroy(alf_handle->threadpool.elems[i].task_thread);
       if(rtn < 0)
       {
            // TODO: need to check if extra things to do in the case of destroy failure
            goto label_thread_mgr_start_err;
       }

       rtn = alf_pal_thread_create(alf_handle->platform_handle, i, 
                              (alf_pal_task_info_t*)&task_handle->task_info, 
#ifdef _ALF_64B_
					  (alf_data_addr64_t)task_handle->p_context,
#else
					  (alf_data_addr64_t)((unsigned int)task_handle->p_context),
#endif
                     &task_handle->p_task_threads[allocated_accels].thread);
       if(rtn < 0)
       {
            // we need to clean up allocated threads so far
            goto label_thread_mgr_start_err;
       }
	     task_handle->p_task_threads[allocated_accels].finished_wbs = 0; // just started
	     task_handle->p_task_threads[allocated_accels].accel_id = i;

	    rtn = alf_thread_mgr_image_info_copy(&alf_handle->threadpool.elems[i], &task_handle->task_info);
      if(rtn < 0)
      {
            // we need to clean up allocated threads so far
            alf_pal_thread_destroy(task_handle->p_task_threads[allocated_accels].thread);
            alf_handle->threadpool.elems[i].status = ALF_API_THREAD_STATUS_INIT;
            goto label_thread_mgr_start_err;
      }
	  
       alf_handle->threadpool.elems[i].task_thread = 
                               task_handle->p_task_threads[allocated_accels].thread;
       alf_handle->threadpool.elems[i].status = ALF_API_THREAD_STATUS_RUNNING;
       alf_handle->threadpool.used_threads ++;
       allocated_accels ++;
       //alf performance hook ALF_THREAD_RUN_INTERVAL begin
       TRACE_INTERVAL_BEGIN(_ALF_THREAD_RUN_INTERVAL,alf_handle->threadpool.elems[i].thread_run_token,1);

       alf_int_task_call_event_handler(task_handle, ALF_TASK_EVENT_INSTANCE_START);
    }

    if(allocated_accels == task_handle->num_accel_req) 
    {
		goto label_thread_mgr_start_done;
    }
  }


#if 0  
  // as designed, we should not have such situation as we have checked before
  // only misuse could cause such problem

  // we can not afford to satsify the task
  if(allocated_accels > 0 && (task_handle->attr & ALF_TASK_ATTR_SCHED_FIXED) ) 
  {
    // clean up the mess :-(
	for(i=0; i<allocated_accels; i++)
	{
        alf_handle->threadpool.elems[task_handle->p_task_threads[i].accel_id].status = ALF_API_THREAD_STATUS_STANDBY;
	}
	task_handle->num_accels = 0;
	alf_handle->threadpool.used_threads -= allocated_accels;
  	rtn = -ALF_ERR_BADR;	// we have no resource
	goto label_thread_mgr_start_err;
  }
#endif

label_thread_mgr_start_done:

  //can be less than required thread number for non-fixed mapping
  task_handle->num_accels = allocated_accels;
  _ALF_DPRINTF(_ALF_ERR_LEVEL_TRACE_, "Entering dataset check\n");
    
  _ALF_DPRINTF(_ALF_ERR_LEVEL_TRACE_, "task_handle->dataset=%p\n",task_handle->dataset);
  
  /* Does this task have an associated data set? */
  if (task_handle->dataset != NULL) {
  	
    _ALF_DPRINTF(_ALF_ERR_LEVEL_TRACE_, "task_handle->dataset != NULL\n");

    /* If so, for each thread (for each accelerator) */
    for (i = 0; i < allocated_accels; i++) {
    	
      _ALF_DPRINTF(_ALF_ERR_LEVEL_TRACE_, "accel %d\n", i);

      /* Associate thread to dataset */
      alf_pal_thread_dataset_associate(task_handle->p_task_threads[i].thread, task_handle->dataset->pal_dataset);
      
    }
    
#ifndef ASSOCIATE_DATASET_TO_ALL_MIDS
    /* Put the PAL dataset to the accelerators */
    alf_pal_dataset_put_no_wait(task_handle->dataset->pal_dataset);
#endif    

    /* Wait for data set to be put into remote data set(s) in accelerator main storage */
    alf_pal_dataset_wait_for_put(task_handle->dataset->pal_dataset);
    
  }

  _ALF_DPRINTF(_ALF_ERR_LEVEL_TRACE_, "Exiting dataset check\n");

  // we have not satisfied the task_handle->num_accel_req, but the task should be ok to run
  // in current situation
  pthread_mutex_unlock(&alf_handle->thread_lock);
  // return the number of allocated accels
  return (int)allocated_accels;

label_thread_mgr_start_err:
  for(i=0; i<allocated_accels; i++)
  {
        alf_handle->threadpool.elems[task_handle->p_task_threads[i].accel_id].status = ALF_API_THREAD_STATUS_STANDBY;
	}
	task_handle->num_accels = 0;
	alf_handle->threadpool.used_threads -= allocated_accels;
  pthread_mutex_unlock(&alf_handle->thread_lock);
	return rtn;
}


/*
 * After task finish its work, the PAL threads will be returned to 
 * thread manager for cache
 */
int
alf_thread_mgr_return_thread(alf_api_t* alf_handle ,alf_api_task_t* task_handle)
{
  unsigned int i;

  // shall we check for errors ?
  pthread_mutex_lock(&alf_handle->thread_lock);
  for(i=0; i<task_handle->num_accels; i++)
  {
	  if(alf_pal_thread_wait(task_handle->p_task_threads[i].thread, 0))
	  {
	  		// need to shutdown it
	  		alf_pal_thread_destroy(task_handle->p_task_threads[i].thread);
		    alf_handle->threadpool.elems[task_handle->p_task_threads[i].accel_id].status = ALF_API_THREAD_STATUS_INIT;
	  }
	  else 
    { // the thread is safe to put back as idle
	  		alf_handle->threadpool.elems[task_handle->p_task_threads[i].accel_id].status = ALF_API_THREAD_STATUS_STANDBY;
    }
    //alf performance hook end
    TRACE_INTERVAL_END(_ALF_THREAD_RUN_INTERVAL,alf_handle->threadpool.elems[task_handle->p_task_threads[i].accel_id].thread_run_token,1,task_handle);
  
    alf_int_task_call_event_handler(task_handle, ALF_TASK_EVENT_INSTANCE_END);
  }
  alf_handle->threadpool.used_threads -= task_handle->num_accels;
  task_handle->num_accels = 0;
  pthread_mutex_unlock(&alf_handle->thread_lock);
  return 0;
}



/*
 * Return how many accelerate is free
 *
 */
int 
alf_thread_mgr_query(alf_api_t* alf_handle)
{
  return alf_handle->threadpool.num_threads - alf_handle->threadpool.used_threads;
}



/*
 * Cleanup the thread pool
 *
 */
int
alf_thread_mgr_cleanup(alf_api_t* alf_handle)
{
  unsigned int i;

  // will not check parm error for internal API

   for(i=0; i<alf_handle->threadpool.num_threads; i++)
  {
    if(alf_handle->threadpool.elems[i].status != ALF_API_THREAD_STATUS_INIT)
    {
      alf_pal_thread_destroy(alf_handle->threadpool.elems[i].task_thread);
    }
  }
  if(alf_handle->threadpool.elems != NULL)
  {
    free(alf_handle->threadpool.elems);
    alf_handle->threadpool.elems = NULL;
  }
  alf_handle->threadpool.num_threads = 0;
  alf_handle->threadpool.used_threads = 0;  // not yet used
  if(alf_handle->accel_num != 0)
    alf_pal_accelerators_release(alf_handle->platform_handle, alf_handle->accel_num);
  alf_handle->accel_num = 0;
  
  return 0;
}

