/* --------------------------------------------------------------  */
/* (C)Copyright 2007                                               */
/* International Business Machines Corporation                     */
/* All Rights Reserved.                                            */
/*                                                                 */
/* Redistribution and use in source and binary forms, with or      */
/* without modification, are permitted provided that the           */
/* following conditions are met:                                   */
/*                                                                 */
/* Redistributions of source code must retain the above copyright  */
/* notice, this list of conditions and the following disclaimer.   */
/*                                                                 */
/* Redistributions in binary form must reproduce the above         */
/* copyright notice, this list of conditions and the following     */
/* disclaimer in the documentation and/or other materials          */
/* provided with the distribution.                                 */
/*                                                                 */
/* Neither the name of IBM Corporation nor the names of its        */
/* contributors may be used to endorse or promote products         */
/* derived from this software without specific prior written       */
/* permission.                                                     */
/*                                                                 */
/* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND          */
/* CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,     */
/* INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF        */
/* MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE        */
/* DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR            */
/* CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,    */
/* SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT    */
/* NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;    */
/* LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)        */
/* HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN       */
/* CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR    */
/* OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,  */
/* EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.              */
/* --------------------------------------------------------------  */

#ifndef __ALF_HOST_H_INC__
#define __ALF_HOST_H_INC__

/*
 ************************************************************
 * ALF Host Code Internal API
 ************************************************************
 */

#include <pthread.h>
#include <stdlib.h>

#include <alf.h>
#include "alf_pal.h"
#include "alf_arraylist.h"
#include "alf_trace.h"

/*
 ************************************************************
 * Adjustable Constants
 ************************************************************
 */

/* Size of the alf handle buffer. It specified the number of alf instances can be created 
 * simultaneously. if only one single instance is expected, set to 1 */
#define ALF_API_ALF_HANDLE_BUFFER_SIZE  (32)

/* Size of the task info handle hash buffer, must be 2^n */
#define ALF_API_TASK_INFO_HASH_BUFFER_SIZE  (256)

/* Size of the task handle hash buffer, must be 2^n */
#define ALF_API_TASK_HANDLE_HASH_BUFFER_SIZE  (1024)

#define _ALF_API_ERR_NUM (128)
#define _ALF_API_TASK_INFO_NUM (128)
#define _ALF_API_TASK_QUEUE_LEN (1024)
#define _ALF_API_TASK_DEP_NUM (8)
#define _ALF_API_TASK_CTX_DESC_NUM (128)
#define _ALF_API_ACCEL_WB_QUEUE_LEN (1024)
#define _ALF_API_ACCEL_DEFAULT_STACK_SIZE (1024)
#define _ALF_API_ACCEL_DEFAULT_ALIGN (7)
#define _ALF_API_WB_DTL_NUM (128)
#define _ALF_API_WB_PARM_MAX_ALIGN (16)
#define _ALF_API_DATASET_LIST_LEN (16)

/*
 ************************************************************
 * Constants
 ************************************************************
 */

typedef enum _ALF_API_TASK_STATE_TYPE_T_ {
  ALF_API_TASK_STATUS_INIT = 0x00000001,
  ALF_API_TASK_STATUS_PENDING = 0x00000002,
  ALF_API_TASK_STATUS_READY = 0x00000003,
  ALF_API_TASK_STATUS_EXEC = 0x00000004,
  ALF_API_TASK_STATUS_FINISH = 0x00000005,
  ALF_API_TASK_STATUS_DESTROY = 0x00000006,
} ALF_API_TASK_STATE_TYPE_T;


typedef enum _ALF_API_TASK_QUERY_STATE_TYPE_T_ {
  ALF_API_TASK_QUERY_STATUS_FINISHED = 0x0,
  ALF_API_TASK_QUERY_STATUS_RUNNING = 0x00000001,
  ALF_API_TASK_QUERY_STATUS_PENDING = 0x00000002,
} ALF_API_TASK_QUERY_STATE_TYPE_T;


typedef enum _ALF_API_STATE_TYPE_T_ {
  ALF_API_STATUS_RUNNING = 0x00000001,
  ALF_API_STATUS_EXITING = 0x00000002,
  ALF_API_STATUS_EXITED = 0x00000003,
} ALF_API_STATE_TYPE_T;

typedef enum _ALF_API_THREAD_STATUS_T_
{
  ALF_API_THREAD_STATUS_INIT     = 0x00000001,   /* thread not created */ 
  ALF_API_THREAD_STATUS_RUNNING  = 0x00000002,   /* thread is running task */
  ALF_API_THREAD_STATUS_STANDBY  = 0x00000003,  /* thread can be reused */
} ALF_API_THREAD_STATUS_T;

typedef enum _ALF_API_WB_STATUS_TYPE_T_ {
  _ALF_API_WB_STATUS_INIT = 0x0,
  _ALF_API_WB_STATUS_ENQ = 0x00000001,
} _ALF_API_WB_STATUS_TYPE_T;

typedef enum _ALF_API_DATASET_STATE_TYPE_T_ {
  ALF_API_DATASET_STATUS_OPEN = 0x00000001,
  ALF_API_DATASET_STATUS_CLOSED = 0x00000002,
  ALF_API_DATASET_STATUS_ERROR = 0x00000003,
} ALF_API_DATASET_STATE_TYPE_T;

/*
 ************************************************************
 * Data Stuctures
 ************************************************************
 */

/* the thread pool instance */
typedef struct _alf_thread_pool_elem
{
  alf_pal_thread_handle task_thread;		// PAL thread handle
  ALF_API_THREAD_STATUS_T status;  			// status of the thread

  // TODO: more smart ways to check image changes
  char lib_name[ALF_STRING_TOKEN_MAX+1];
  char image_name[ALF_STRING_TOKEN_MAX+1];
  TRACE_INTERVAL_TOKEN_ARGUMENT(thread_run_token);
} alf_thread_pool_elem_t;


/* the thread pool */
typedef struct _alf_thread_pool
{
  alf_thread_pool_elem_t* elems;	//  threads
  unsigned int num_threads;      	// number of threads in pool
  unsigned int used_threads;			// count of threads in use
} alf_thread_pool;




/* alf_handle_t internal data structure  */
typedef struct _alf_api_t_ {
  pthread_mutex_t lock;         // exclusive lock
  pthread_mutex_t cond_lock;    //cond lock
  pthread_cond_t cond;
  alf_handle_t self;
  unsigned int flag;

  volatile ALF_API_STATE_TYPE_T state;  //alf runtime state

  alf_thread_pool threadpool;   // the thread pool of this alf instance
  pthread_mutex_t thread_lock;
  unsigned int max_thread_num;   //maximum of threads number supported
  unsigned int accel_num;       // number of accelerators reserved currently

  unsigned int task_unfinished; //number of unfinished task when exit

  alf_arraylist_t *init_task_list;
  alf_arraylist_t *ready_task_list;
  alf_arraylist_t *exec_task_list;
  alf_arraylist_t *destroy_task_list;
  alf_arraylist_t *ref_task_list;
  
  alf_arraylist_t *task_info_list;

  alf_pal_handle platform_handle;

  // error handler
  alf_error_handler_t err_handler;
  void *err_handler_data_ptr;
  alf_arraylist_t *err_msg_list;
  pthread_mutex_t err_lock;
  pthread_t err_thread;
  volatile unsigned long long err_num;
  volatile unsigned long long err_finished;

  // scheduler thread
  pthread_t scheduler;
  ALF_EXIT_POLICY_T exit_policy;

  alf_arraylist_t *datasets;    // list of data sets

} alf_api_t;

typedef struct _alf_err_msg {
  int error_type;
  ALF_ERR_TYPE_T error_code;
  char *error_string;
  ALF_ERR_POLICY_T ret;
} alf_err_msg;

/* the alf data set */
typedef struct _alf_api_dataset_t_ {
  pthread_mutex_t lock;
  alf_api_t *api_handle;        // backward ptr to api runtime object owning this data set
  ALF_API_DATASET_STATE_TYPE_T state;   // state of this data set
  alf_arraylist_t *buffers;     // list of data set buffers in this data set
  alf_arraylist_t *tasks;       // list of tasks associated to this data set
  alf_pal_dataset_handle pal_dataset;   // PAL data set
} alf_api_dataset_t;

/* the alf data set buffer */
typedef struct _alf_api_dataset_buffer_t_ {
  alf_data_addr64_t addr;       // data buffer address
  unsigned long long size;      // data size
  ALF_DATASET_ACCESS_MODE_T access_mode;        // data access mode
} alf_api_dataset_buffer_t;

/* alf_task_info_handle_t internal data structure */

typedef struct _alf_api_task_info_t_ {
  alf_pal_task_info_t pal;      /* this must be the first field */
  pthread_mutex_t lock;
  alf_api_t *alf_handle;
  /* host data structure */
  alf_task_handle_t self;       // self pointer
  /* hook to next in the hash bucket */
  struct _alf_api_task_info_t_ *next_ptr;
  unsigned int accel_mem_size;
  unsigned int accel_align;
  unsigned int max_context_desc_num;
  unsigned int ctx_entry_size;

  /* Fixme:  other members TBD ? */

} alf_api_task_info_t;


/* thread of a task */
typedef struct _alf_task_thread
{
  alf_pal_thread_handle thread;  // PAL thread
  int		accel_id;
  unsigned int finished_wbs;
} alf_task_thread_t;


typedef struct _alf_task_event_handler
{
  unsigned int event_mask;
  int (*callback_func)(alf_task_handle_t task_handle, ALF_TASK_EVENT_TYPE_T event,
                            void* p_data);
  void *p_data;
  unsigned int data_size;
} alf_task_event_handler_t;

/* alf_task_handle_t internal data structure */

typedef struct _alf_api_task_t_ {
  /* host data structure */
  alf_task_handle_t self;       // self pointer
  alf_api_t *alf_handle;		// refer to alf handle
  
  struct _alf_api_task_t_ *next_ptr;    /* TBD: hook to next in the hash bucket */

  // synchronization flags
  pthread_mutex_t lock;         // exclusive lock
  pthread_mutex_t cond_lock;    //cond lock
  pthread_cond_t cond;


  /* basic task enviromental information */
  alf_api_task_info_t task_info;        /* a copied task info that is referenced by this task only */

  /* all other static task information follow here */
  unsigned int attr;
  unsigned int wb_dist_size;
  unsigned int num_accel_req;   // required number of accelerators

  alf_task_event_handler_t event_handler;

  void *p_context;   // context buffer as supplied

  /* task status info */
  ALF_API_TASK_STATE_TYPE_T state;
  unsigned int finalized_flag;
  unsigned int status;

  alf_task_thread_t *p_task_threads;        /* the task thread container */
  void *p_thread_context;
  unsigned int max_desc_num;
  unsigned int num_accels;      /* actually utilized number of accelerators */

  alf_arraylist_t *task_wbq;
  alf_arraylist_t **accel_wbq;
  alf_arraylist_t **accel_fin_wbq;

  /* record dependency */
  unsigned int parent_count;
  alf_arraylist_t* child_tasks;

  unsigned int dist_num;
  int cur_accel;


  volatile int num_wb_pending;
  volatile int num_wb;
  /* Fixme:  any other TBDs */

  alf_api_dataset_t *dataset;   // associated data set

  TRACE_INTERVAL_TOKEN_ARGUMENT(task_before_exec_token);
  TRACE_INTERVAL_TOKEN_ARGUMENT(task_in_exec_token);

} alf_api_task_t;


/* alf_wb_handle_t internal data structure */

typedef struct _alf_api_wb_t_ {
  alf_pal_wb_t pal;             /* this must be the first field */

  alf_api_task_t *task;         // associated task handle
  
  _ALF_API_WB_STATUS_TYPE_T status;

  alf_pal_dtl_handle cur_dtl;   // used for current open DTL
  unsigned int cur_dtl_size;
  unsigned int cur_dtl_offset;
  ALF_BUF_TYPE_T cur_dtl_buffer_type; 


} alf_api_wb_t;


/* the alf handle hash table */
typedef struct _alf_api_task_info_bucket_t_ {
  pthread_mutex_t lock;         // exclusive lock
  unsigned int count;           // current bucket depth
  alf_api_task_info_t *top_ptr; // the first one in bucket
} alf_api_task_info_bucket_t;


/* the alf handle hash table */
typedef struct _alf_api_task_bucket_t_ {
  pthread_mutex_t lock;         // exclusive lock
  unsigned int count;           // current bucket depth
  alf_api_task_t *top_ptr;      // the first one in bucket
} alf_api_task_bucket_t;


#ifdef __cplusplus
extern "C" {
#endif                          /* __cplusplus */

/*
 ************************************************************
 * APIs
 ************************************************************
 */

/* handle operations */
#define ALF_API_GET_ALF_HANDLE(handle) ((alf_data_uint32_t)((handle)>>32))

#define ALF_API_GET_HANDLE_HASH(handle) ((alf_data_uint32_t)(handle))

#define ALF_API_MAKE_HANDLE(alf_handle, handle) ( ((alf_data_uint64)(alf_handle)<<32) \
                                                     |(alf_data_uint64)(handle))

  /* hash table */
#define ALF_API_TASK_INFO_HASH_INDEX(handle_index)     ((handle_index&(ALF_API_TASK_INFO_HASH_BUFFER_SIZE-1))
#define ALF_API_TASK_HASH_INDEX(handle_index)     ((handle_index)&(ALF_API_TASK_HANDLE_HASH_BUFFER_SIZE-1))


/* host internal api with prefix alf_api_ */
  void *alf_api_scheduler(void *alf_ptr);
  void *alf_api_err_thread(void *alf_ptr);
  int alf_api_task_wb_enqueue(alf_api_task_t * task_handle, alf_api_wb_t * wb_handle);
  void alf_api_handle_destroy(alf_api_t * alf_handle);
  int alf_api_task_offspring_destroy(alf_api_task_t *task_handle);
  int alf_api_task_info_copy(alf_api_task_t* task_handle, alf_api_task_info_t* task_info, 
                       void *p_task_context_data);
  int alf_api_task_wbq_create (alf_api_task_t * task_handle);

/* error handler internal api with prefix alf_err_ */
  ALF_ERR_POLICY_T  alf_err_default_error_handler(void *p_context_data  DECLARE_UNUSED, ALF_ERR_TYPE_T error_type, int error_code, char *error_string);
  ALF_ERR_POLICY_T  alf_err_pal_error_handler(void *p_context_data, ALF_ERR_TYPE_T error_type, int error_code, char *error_string);
  ALF_ERR_POLICY_T alf_err_call_error_handler(alf_api_t* alf_handle, ALF_ERR_TYPE_T error_type, int error_code, char *err_msg);
  ALF_ERR_POLICY_T alf_err_error_processing(alf_api_t* alf_handle, ALF_ERR_TYPE_T error_type, int error_code, char *err_msg);

/* host and scheduler shared internal api with prefix alf_int_ */
  void alf_int_task_info_destroy(alf_api_task_info_t * task_info);
  void alf_int_task_res_destroy(alf_api_task_t * task_handle);
  void alf_int_wb_handle_destroy(alf_api_wb_t * wb);
  int alf_int_task_call_event_handler(alf_api_task_t* task_handle, ALF_TASK_EVENT_TYPE_T event);

/* scheduler internal api with prefix alf_sched_ */
  void alf_sched_task_wb_count_dec(alf_api_task_t * task_handle);
  int alf_sched_task_wb_dispatch(alf_api_task_t * task_handle);
  void alf_sched_task_wb_try_free(alf_api_task_t * task_handle);
  void alf_sched_task_wb_free(alf_api_task_t * task_handle);
  int alf_sched_task_wbq_create (alf_api_task_t * task_handle);
  int alf_sched_task_cancel(alf_api_t * alf_handle);
  int alf_sched_task_release (alf_api_t* alf_handle);
  void alf_sched_task_select(alf_api_t * alf_handle);
  void alf_sched_task_schedule(alf_api_t* alf_handle);
  void alf_sched_task_run(alf_api_t* alf_handle);
  void alf_sched_task_context_merge(alf_api_task_t *task_handle, unsigned int depth);
  void alf_sched_task_child_notify(alf_api_task_t* task_handle);

/* thread manager internal api with prefix alf_thread_mgr_ */
int alf_thread_mgr_setup(alf_api_t* alf_handle, unsigned int num_of_accels);
int alf_thread_mgr_start_thread(alf_api_t* alf_handle, alf_api_task_t* task_handle);
int alf_thread_mgr_return_thread(alf_api_t* alf_handle ,alf_api_task_t* task_handle);
int alf_thread_mgr_query(alf_api_t* alf_handle);
int alf_thread_mgr_cleanup(alf_api_t* alf_handle);

void alf_api_task_dump(char *prefix, alf_api_task_t * wb);
void alf_api_wb_dump(char *prefix, alf_api_wb_t * wb);
void alf_api_dataset_dump(char *prefix, alf_api_dataset_t * dataset);

  int alf_api_dataset_check(alf_api_dataset_t * dataset, alf_data_addr64_t addr, unsigned long long size,
                            ALF_DATASET_ACCESS_MODE_T access_mode);
  int alf_api_dataset_put(alf_api_dataset_t * dataset);
  int alf_api_dataset_get(alf_api_dataset_t * dataset);
  int alf_api_dataset_close(alf_api_dataset_t * dataset);
/*
 ************************************************************
 * Globals
 ************************************************************
 */


#ifdef __cplusplus
}
#endif                          /* __cplusplus */
#endif                          /* __ALF_HOST_H_INC__ */
